#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the MouseInteractorStyle class.
"""

from uuid import uuid1
from enum import Enum, auto
from typing import cast, Any, Optional, Callable, Dict, Tuple, Set, List

from vtk import (
    vtkActor,
    vtkRenderer,
    vtkIdTypeArray,
    vtkMolecule,
    vtkMoleculeMapper,
    vtkHardwareSelector,
    vtkInteractorStyleTrackballCamera,
)

import scine_utilities as su

from scine_heron.electronic_data.electronic_data_status_manager import (
    ElectronicDataStatusManager,
)
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from scine_heron.molecule.utils.molecule_utils import molecule_to_atom_collection
from scine_heron.settings.settings_status_manager import SettingsStatusManager


class InteractionMode(Enum):
    NO_INTERACTION = auto()
    ROTATE_CAMERA = auto()
    MOVE_CAMERA = auto()
    MOVE_ATOMS = auto()


class MouseInteractorStyle(vtkInteractorStyleTrackballCamera):  # type: ignore[misc]
    """
    An interactor style (based on "vtkInteractorStyleTrackballCamera") that supports
    picking and moving particles with the right mouse button.
    """

    def __init__(
        self,
        interactor: QVTKRenderWindowInteractor,
        renderer: vtkRenderer,
        mapper: vtkMoleculeMapper,
        actors_dict: Dict[str, vtkActor],
        selected_atom_callback: Callable[[Optional[List[int]]], None],
    ):
        vtkInteractorStyleTrackballCamera.__init__(self)

        self.__interactor = interactor
        self.__mapper = mapper
        self.__actors_dict = actors_dict

        self._renderer = renderer
        self._selected_atom_callback = selected_atom_callback
        self._pressed_buttons: Set[str] = set()

        self._settings_status_manager: SettingsStatusManager = None  # type: ignore[assignment]

        self._electronic_data_status_manager: Optional[
            ElectronicDataStatusManager
        ] = None

        self._molecule = vtkMolecule()
        self._molecule_version = uuid1()

        self.__add_observers()

    @staticmethod
    def _interaction_mode(
        pressed_buttons: Set[str], selected_atom_ids: Optional[List[int]]
    ) -> InteractionMode:
        if "left_mouse" in pressed_buttons and " " not in pressed_buttons:
            mode = InteractionMode.ROTATE_CAMERA
        elif "left_mouse" in pressed_buttons and " " in pressed_buttons:
            mode = InteractionMode.MOVE_CAMERA
        elif "right_mouse" in pressed_buttons and selected_atom_ids:
            mode = InteractionMode.MOVE_ATOMS
        elif "right_haptic" in pressed_buttons and selected_atom_ids:
            mode = InteractionMode.MOVE_ATOMS
        else:
            mode = InteractionMode.NO_INTERACTION

        return mode

    def set_settings_status_manager(
        self, settings_status_manager: SettingsStatusManager
    ) -> None:
        self._settings_status_manager = settings_status_manager

    def set_electronic_data_status_manager(
        self, electronic_data_status_manager: ElectronicDataStatusManager
    ) -> None:
        self._electronic_data_status_manager = electronic_data_status_manager

    def __add_observers(self) -> None:
        """
        Add observers for mouse presses and movements.
        """
        priority = 1.0
        self.AddObserver(
            "LeftButtonPressEvent", self._handle_left_button_press, priority
        )
        self.AddObserver(
            "LeftButtonReleaseEvent", self._handle_left_button_release, priority
        )
        self.AddObserver(
            "RightButtonPressEvent", self.__handle_right_button_press, priority
        )
        self.AddObserver(
            "RightButtonReleaseEvent", self.__handle_right_button_release, priority
        )
        self.AddObserver("MouseMoveEvent", self.__handle_mouse_move, priority)
        self.AddObserver("KeyPressEvent", self.__key_press, priority)
        self.AddObserver("KeyReleaseEvent", self.__key_release, priority)
        self.AddObserver(
            "MouseWheelForwardEvent", self._handle_mouse_wheel_forward, priority
        )
        self.AddObserver(
            "MouseWheelBackwardEvent", self._handle_mouse_wheel_backward, priority
        )

    def _handle_mouse_wheel_forward(self, _1: Optional[Any] = None, _2: Optional[Any] = None) -> None:
        self.OnMouseWheelForward()
        self._render()

    def _handle_mouse_wheel_backward(self, _1: Optional[Any] = None, _2: Optional[Any] = None) -> None:
        self.OnMouseWheelBackward()
        self._render()

    def _handle_left_button_press(self, _1: Optional[Any] = None, _2: Optional[Any] = None) -> None:
        """
        Activates "rotation mode".
        """
        self._pressed_buttons.add("left_mouse")

    def _handle_left_button_release(self, _1: Optional[Any] = None, _2: Optional[Any] = None) -> None:
        """
        Deactivates "rotation mode".
        """
        self._pressed_buttons.remove("left_mouse")

    def __handle_right_button_press(self, _1: Optional[Any] = None, _2: Optional[Any] = None) -> None:
        """
        Sets the member mouse_picked_atom_ids to the picked atoms.
        """
        self._pressed_buttons.add("right_mouse")
        self._settings_status_manager.mouse_picked_atom_ids = self.__picked_atoms(
            self._settings_status_manager.mouse_picked_atom_ids)
        self._selected_atom_callback(self._settings_status_manager.mouse_picked_atom_ids)

    def __handle_right_button_release(self, _1: Optional[Any] = None, _2: Optional[Any] = None) -> None:
        """
        Resets the member mouse_picked_atom_ids to None.
        """
        self.__key_press()
        self._pressed_buttons.remove("right_mouse")
        if (
            self._settings_status_manager.mouse_picked_atom_ids
            and len(self._settings_status_manager.mouse_picked_atom_ids) == 1
            and "s" not in self._pressed_buttons
        ):
            self._settings_status_manager.mouse_picked_atom_ids = []

    def __handle_mouse_move(self, _1: Any, _2: Any) -> None:
        """
        Moves the picked atoms to the location defined by the mouse position.
        """
        interaction_mode = self._interaction_mode(
            self._pressed_buttons, self._settings_status_manager.mouse_picked_atom_ids
        )

        last_xy_pos = self.GetInteractor().GetLastEventPosition()
        xy_pos = self.GetInteractor().GetEventPosition()

        if interaction_mode == InteractionMode.ROTATE_CAMERA:
            self._rotate_camera(last_xy_pos[0] - xy_pos[0], last_xy_pos[1] - xy_pos[1])
            self._render()
        elif interaction_mode == InteractionMode.MOVE_CAMERA:
            self._move_camera(self.__calculate_motion_vector(last_xy_pos, xy_pos))
        elif interaction_mode == InteractionMode.MOVE_ATOMS:
            self._settings_status_manager.selected_molecular_orbital = None
            self._settings_status_manager.number_of_molecular_orbital = None

            self._move_atoms_by_mouse()
            self._render()
        else:
            self.OnMouseMove()

    def __calculate_motion_vector(
        self, last_xy_pos: Tuple[float, float], xy_pos: Tuple[float, float]
    ) -> Tuple[float, float, float]:
        last_xy_pos_world = self.__view_to_world(self.__display_to_view(last_xy_pos))
        xy_pos_world = self.__view_to_world(self.__display_to_view(xy_pos))

        return (
            last_xy_pos_world[0] - xy_pos_world[0],
            last_xy_pos_world[1] - xy_pos_world[1],
            last_xy_pos_world[2] - xy_pos_world[2],
        )

    def _move_camera(self, motion_vector: Tuple[float, float, float]) -> None:
        camera = self._renderer.GetActiveCamera()
        focal = camera.GetFocalPoint()
        point = camera.GetPosition()

        camera.SetFocalPoint(
            motion_vector[0] + focal[0],
            motion_vector[1] + focal[1],
            motion_vector[2] + focal[2],
        )
        camera.SetPosition(
            motion_vector[0] + point[0],
            motion_vector[1] + point[1],
            motion_vector[2] + point[2],
        )
        camera.OrthogonalizeViewUp()

        self._render()

    def _rotate_camera(self, azimuth: Any, elevation: Any) -> None:
        camera = self._renderer.GetActiveCamera()
        atoms = molecule_to_atom_collection(self._molecule)
        com = su.geometry.get_center_of_mass(atoms) * su.ANGSTROM_PER_BOHR
        camera.SetFocalPoint(com)
        camera.OrthogonalizeViewUp()
        camera.Azimuth(azimuth)
        camera.Elevation(elevation)

    def _move_atoms_by_mouse(self) -> None:
        # get atom position
        if self._settings_status_manager.mouse_picked_atom_ids:
            for atom_id in self._settings_status_manager.mouse_picked_atom_ids:
                atom = self._molecule.GetAtom(atom_id)
                from_view = self.__world_to_view(atom.GetPosition())

            for this in self._settings_status_manager.mouse_picked_atom_ids:
                atom = self._molecule.GetAtom(this)
                # get mouse position and convert it to view coordinates
                to_view = self.__display_to_view(self.GetInteractor().GetEventPosition())
                from_view = self.__world_to_view(atom.GetPosition())
                to_world = self.__view_to_world((to_view[0], to_view[1], from_view[2]))
                atom.SetPosition(to_world[0], to_world[1], to_world[2])

                for other in self._settings_status_manager.mouse_picked_atom_ids:
                    if other is not this:
                        otherAtom = self._molecule.GetAtom(other)
                        from_view_other = self.__world_to_view(otherAtom.GetPosition())
                        dx = (from_view_other[0] - from_view[0])
                        dy = (from_view_other[1] - from_view[1])
                        to_world_other = self.__view_to_world((to_view[0] + dx, to_view[1] + dy, from_view_other[2]))
                        otherAtom.SetPosition(to_world_other[0], to_world_other[1], to_world_other[2])

            # break bonds without running calculation
            bonds = self._molecule.GetNumberOfBonds()
            for bond in range(bonds):
                if self._molecule.GetBondLength(bond) > 1.0:
                    self._molecule.SetBondOrder(bond, 0)

    def __display_to_view(
        self, display: Tuple[float, float]
    ) -> Tuple[float, float, float]:
        """
        Converts display coordinates to view coordinates.
        """
        self._renderer.SetDisplayPoint(display[0], display[1], 0)
        self._renderer.DisplayToView()
        return cast(Tuple[float, float, float], self._renderer.GetViewPoint())

    def __world_to_view(
        self, world: Tuple[float, float, float]
    ) -> Tuple[float, float, float]:
        """
        Converts world to view coordinates.
        """
        self._renderer.SetWorldPoint(world[0], world[1], world[2], 1.0)
        self._renderer.WorldToView()
        return cast(Tuple[float, float, float], self._renderer.GetViewPoint())

    def __view_to_world(
        self, view: Tuple[float, float, float]
    ) -> Tuple[float, float, float]:
        """
        Converts view to world coordinates.
        """
        self._renderer.SetViewPoint(view[0], view[1], view[2])
        self._renderer.ViewToWorld()
        return cast(Tuple[float, float, float], self._renderer.GetWorldPoint())

    def __key_press(self, _1: Optional[Any] = None, _2: Optional[Any] = None) -> None:
        """
        Translate camera if the space bar has been pressed.
        """
        self._pressed_buttons.add(self.__interactor.GetKeyCode())

    def __key_release(self, _1: Optional[Any] = None, _2: Optional[Any] = None) -> None:
        """
        Stop translating camera if the space bar has been released.
        """
        key = self.__interactor.GetKeyCode()
        if key in self._pressed_buttons:
            self._pressed_buttons.remove(self.__interactor.GetKeyCode())

    def __picked_atoms(self, picked_atoms_list: Optional[List[int]]) -> Optional[List[int]]:
        """
        Uses the event position to return the picked atoms or None.
        """
        from scine_heron.molecular_viewer import get_mol_viewer_tab
        self.__key_press()
        mol_tab = get_mol_viewer_tab(want_atoms_there=True)

        if not picked_atoms_list:
            picked_atoms_list = []

        selector = vtkHardwareSelector()
        selector.SetRenderer(self._renderer)
        position = self.GetInteractor().GetEventPosition()

        selector.SetArea(position[0], position[1], position[0], position[1])

        # before we retrieve the selection, we hide all actors but the molecule actor
        # this allows the selector to correctly select the atom that was picked
        # removing the actors caused problems with multiple active selections, but hiding looks to be fine
        actors_2d = [a for a in self._renderer.GetActors2D()]
        actors_3d = [a for a in self._renderer.GetActors()
                     if a != self.__actors_dict["molecule"]]
        all_actors = actors_2d + actors_3d
        for actor in all_actors:
            actor.SetVisibility(False)
        selection = selector.Select()
        for actor in all_actors:
            actor.SetVisibility(True)
        ids = vtkIdTypeArray()  # ids are written into here from mapper
        self.__mapper.GetSelectedAtoms(selection, ids)

        if ids.GetSize() > 0 and int(ids.GetValue(0)) not in picked_atoms_list:
            if "s" in self._pressed_buttons:
                picked_atoms_list.append(int(ids.GetValue(0)))
            else:
                picked_atoms_list = [int(ids.GetValue(0))]
        elif (
            ids.GetSize() > 0
            and int(ids.GetValue(0)) in picked_atoms_list
            and "s" in self._pressed_buttons
            and len(picked_atoms_list) > 1
        ):
            picked_atoms_list.remove(int(ids.GetValue(0)))
        if mol_tab and mol_tab.mol_widget:
            mol_tab.mol_widget.set_selection(picked_atoms_list)
        return picked_atoms_list

    def _render(self) -> None:
        """
        Render the scene.
        """
        self._renderer.GetRenderWindow().Render()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
import json
import numpy as np
from typing import Any, Optional, List, Dict, Tuple, Union

import scine_utilities as utils
import scine_database as db
from scine_database.energy_query_functions import (
    get_barriers_for_elementary_step_by_type,
    get_elementary_step_with_min_ts_energy
)
from scine_heron.database.graphics_items import Compound, Reaction
from scine_heron.molecule.molecule_video import MoleculeVideo
from scine_heron.molecule.molecule_widget import MoleculeWidget
from scine_heron.molecule.reaction_profile import ReactionProfileWidget
from scine_heron.utilities import (
    copy_text_to_clipboard,
    qcolor_by_key,
    build_brush,
    build_pen,
)

from PySide2.QtWidgets import (
    QGraphicsItem,
    QGraphicsPathItem,
    QGraphicsView,
    QWidget,
    QGraphicsScene,
    QLayout,
    QMenu,
    QFileDialog
)
from PySide2.QtGui import (
    QKeySequence,
    QGuiApplication,
    QPainter,
)
from PySide2.QtSvg import (
    QSvgGenerator,
)
from PySide2.QtCore import (
    Qt,
    QEvent,
    QSize
)


class ReactionAndCompoundViewSettings(QWidget):
    def __init__(self, parent: QWidget, parsed_layout: QLayout):
        super(ReactionAndCompoundViewSettings, self).__init__(parent)
        self.p_layout = parsed_layout
        self.mol_widget_cache = MoleculeWidget(
            parent=self,
            alternative_zoom_controls=True,
            disable_modification=True
        )
        self.mol_widget = self.mol_widget_cache
        self.es_mep_widget = ReactionProfileWidget(parent=self)

    def update_molecule_widget(self, new_widget):
        self.p_layout.replaceWidget(self.mol_widget, new_widget)
        if isinstance(self.mol_widget, MoleculeVideo):
            self.mol_widget.close()
        self.mol_widget = new_widget
        self.mol_widget.show()


class ReactionAndCompoundView(QGraphicsView):
    def __init__(self, parent: QWidget, width: Optional[int] = None, height: Optional[int] = None):

        if width and height:
            self.scene_object = QGraphicsScene(0, 0, width, height)
        else:
            self.scene_object = QGraphicsScene()

        super(ReactionAndCompoundView, self).__init__(self.scene_object, parent=parent)
        self.settings: ReactionAndCompoundViewSettings

        # settings regarding presentation
        self.compound_color = qcolor_by_key('compoundColor')
        self.gray_out_color = qcolor_by_key('grayOutColor')
        self.reaction_color = qcolor_by_key('reactionColor')
        self.highlight_color = qcolor_by_key('highlightColor')
        self.border_color = qcolor_by_key('borderColor')
        self.border_gray_color = qcolor_by_key('borderGrayColor')
        self.edge_color = qcolor_by_key('edgeColor')
        self.association_color = qcolor_by_key('associationColor')
        self.flask_color = qcolor_by_key('flaskColor')

        # rendering smoother lines and edges of nodes
        self.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)

        self.reaction_brush = build_brush(self.reaction_color)
        self.reaction_pen = build_pen(self.border_color)
        self.compound_brush = build_brush(self.compound_color)
        self.compound_pen = build_pen(self.border_color)
        self.gray_border_pen = build_pen(self.border_gray_color)
        self.flask_brush = build_brush(self.flask_color)
        self.association_brush = build_brush(self.association_color)
        self.hover_brush = build_brush(self.highlight_color)
        self.hover_pen = build_pen(self.highlight_color, width=2)
        self.gray_out_pen = build_pen(self.gray_out_color)
        self.path_pen = build_pen(self.edge_color)

        self.focused_item_db_id: Optional[str] = None
        self.focused_item: Optional[QGraphicsItem] = None
        self.focused_connected_items: List[str] = []
        self.focused_connected_lines: List[str] = []

        self.view_highlighted = False

        self.compounds: Dict[str, List[Compound]] = dict()
        self.reactions: Dict[str, List[Reaction]] = dict()
        self.line_items: Dict[str, QGraphicsPathItem] = dict()

        self.__expanded_views: List[QWidget] = []

    def save_svg(self):
        filename, _ = QFileDialog.getSaveFileName(
            self,
            self.tr("Save File"),  # type: ignore[arg-type]
            "network.svg",
            self.tr("Vector Graphics (*.svg)"),  # type: ignore[arg-type]
        )
        generator = QSvgGenerator()
        generator.setFileName(filename)
        generator.setSize(QSize(self.sceneRect().width(), self.sceneRect().height()))
        painter = QPainter()
        painter.begin(generator)
        self.scene_object.render(painter)
        painter.end()

    def move_to_foreground(self, dictionary: Dict[str, List[QGraphicsItem]]) -> None:
        for k in dictionary.values():
            for item in k:
                self.scene_object.removeItem(item)
                self.scene_object.addItem(item)

    def replace_in_compound_list(self, item: Compound):
        str_id = item.db_representation.id().string()
        self.compounds[str_id] = [item]

    def add_to_compound_list(self, item: Compound):
        str_id = item.db_representation.id().string()
        if str_id in self.compounds:
            self.compounds[str_id].append(item)
        else:
            self.compounds[str_id] = [item]

    def add_to_reaction_list(self, item: Reaction):
        str_id = item.db_representation.id().string()
        if str_id in self.reactions:
            self.reactions[str_id].append(item)
        else:
            self.reactions[str_id] = [item]

    def set_settings_widget(self, settings_widget: ReactionAndCompoundViewSettings):
        self.settings = settings_widget

    @staticmethod
    def __spline_to_trajectory(spline: utils.bsplines.TrajectorySpline, n_frames: int = 50) \
            -> utils.MolecularTrajectory:
        trajectory = utils.MolecularTrajectory()
        _, atoms = spline.evaluate(0, 3)
        trajectory.elements = atoms.elements
        for i in range(n_frames):
            p = i / n_frames
            energy, atoms = spline.evaluate(p, 3)
            trajectory.push_back(atoms.positions, energy)
        return trajectory

    def get_aggregate_brush(self, a_type: db.CompoundOrFlask):
        if a_type == db.CompoundOrFlask.FLASK:
            return self.flask_brush
        return self.compound_brush

    def get_reaction_brush(self, elementary_step_type: db.ElementaryStepType):
        if elementary_step_type == db.ElementaryStepType.BARRIERLESS:
            return self.association_brush
        return self.reaction_brush

    def wheelEvent(self, event):
        modifiers = QGuiApplication.keyboardModifiers()
        if modifiers == Qt.ControlModifier:
            zoomInFactor = 1.25
            zoomOutFactor = 1 / zoomInFactor

            # Save the scene pos
            oldPos = self.mapToScene(event.pos())

            # Zoom
            if event.angleDelta().y() > 0:
                zoomFactor = zoomInFactor
            else:
                zoomFactor = zoomOutFactor
            self.scale(zoomFactor, zoomFactor)

            # Get the new position
            newPos = self.mapToScene(event.pos())

            # Move scene to old position
            delta = newPos - oldPos
            self.translate(delta.x(), delta.y())
        elif event.modifiers() == Qt.ShiftModifier:
            self.horizontalScrollBar().wheelEvent(event)
        else:
            self.verticalScrollBar().wheelEvent(event)

    def mousePressEvent(self, ev):
        if ev.button() == Qt.LeftButton:
            if not isinstance(self.itemAt(ev.pos()), Compound) and not isinstance(self.itemAt(ev.pos()), Reaction):
                self.reset_node_focus()
        super().mousePressEvent(ev)

    def reset_node_focus(self):
        self.reset_item_colors()
        # NOTE: Guess this is somewhat causing the weird blink
        self.settings.es_mep_widget.clear_canvas()

    def keyPressEvent(self, event):
        if (event.type() == QEvent.KeyPress and event == QKeySequence.Copy):
            clipboard = QGuiApplication.clipboard()
            if self.focused_item_db_id is not None and \
                    self.focused_item_db_id not in list(self.reactions.keys()):
                text = self.focused_item_db_id
            elif len(self.compounds) != 0:
                text = list(self.compounds.keys())[0]
            else:
                text = ""
            clipboard.setText(text)

    def reset_item_colors(self):
        # # # Reset all compounds
        for c_list in self.compounds.values():
            for entry in c_list:
                entry.reset_brush()
        # # # Reset all reactions
        for r_list in self.reactions.values():
            for entry in r_list:
                entry.reset_brush()
        # # # Reset all lines
        for line in self.line_items.values():
            line.setPen(self.path_pen)

        self.view_highlighted = False

    # NOTE: Ideally, this should be derived from the subgraph of a centroid; would require the cache to be attribute of
    # this class.
    def set_brush_pen_for_connected_items(self, str_id: str):
        # # # Derive connected items and lines, would be better with graph
        self.focused_connected_items = [str_id]
        self.focused_connected_lines = []
        for edge_key in self.line_items.keys():
            if str_id in edge_key:
                self.focused_connected_items += [c_item[0:24] for c_item in edge_key.split("_")[0:2]
                                                 if str_id != c_item[0:24]]
                self.focused_connected_lines.append(edge_key)

        for item_key in self.compounds.keys():
            if item_key in self.focused_connected_items:
                self.scene_object.removeItem(self.compounds[item_key][0])
                self.scene_object.addItem(self.compounds[item_key][0])
                continue
            else:
                entry = self.compounds[item_key]
                for item in entry:
                    item.set_current_brush(self.gray_out_color)
                    item.set_current_pen(self.gray_border_pen)

        for item_key in self.reactions.keys():
            if item_key in self.focused_connected_items:
                self.scene_object.removeItem(self.reactions[item_key][0])
                self.scene_object.addItem(self.reactions[item_key][0])
                continue
            else:
                entry = self.reactions[item_key]
                for item in entry:
                    item.set_current_brush(self.gray_out_color)
                    item.set_current_pen(self.gray_border_pen)

        for line_key in self.line_items.keys():
            if line_key in self.focused_connected_lines:
                continue
            else:
                self.line_items[line_key].setPen(self.gray_out_pen)

        self.view_highlighted = True

    def set_hover(self, item: QGraphicsItem) -> None:
        item_db_id_h = item.db_representation.id().string()
        item.setPen(self.hover_pen)
        item.setBrush(self.hover_brush)

        for k in self.line_items.keys():
            if item_db_id_h in k:
                self.line_items[k].setPen(self.hover_pen)

    def reset_hover(self, item: QGraphicsItem) -> None:
        item_db_id = item.db_representation.id().string()
        # # # Choose correct brush during hover
        if self.view_highlighted and item_db_id not in self.focused_connected_items:
            line_pen = self.gray_out_pen
            border_pen = self.gray_border_pen
            item.set_current_brush(self.gray_out_color)
            item.set_current_pen(self.gray_border_pen)
        else:
            line_pen = self.path_pen
            border_pen = self.compound_pen if isinstance(item, Compound) else self.reaction_pen
            item.reset_brush()

        if item_db_id == self.focused_item_db_id:
            border_pen = self.hover_pen
        item.set_current_pen(border_pen)
        for k in self.line_items.keys():
            if item_db_id in k:
                self.line_items[k].setPen(line_pen)

    def mouse_press_function(self, _, item: QGraphicsItem) -> None:
        if not self.settings:
            raise RuntimeError('Settings were never added to the reaction compound widget.')
        # update widget
        self.focused_item = item
        if item.db_representation is not None:
            self.focused_item_db_id = str(item.db_representation.id())
        trajectory: Any = utils.MolecularTrajectory()

        mol_widget = self.settings.mol_widget_cache
        mol_index = self.settings.p_layout.indexOf(self.settings.mol_widget)
        charge = None
        mult = None
        if isinstance(item, Reaction):
            if item.spline is not None:
                trajectory = self.__spline_to_trajectory(item.spline)
                # invert trajectory if required
                if item.invert_direction:
                    trajectory = trajectory[::-1]
                # TODO: Avoid database call here!
                step = db.ElementaryStep(item.assigned_es_id, self._elementary_step_collection)
                ts = db.Structure(step.get_transition_state(), self._structure_collection)
                charge = ts.get_charge()
                mult = ts.get_multiplicity()
            else:
                # barrierless reaction currently no trajectory info and no static structure to display
                s = utils.AtomCollection()
                trajectory.elements = s.elements
                trajectory.push_back(s.positions)

            new_widget = MoleculeVideo(parent=self, trajectory=trajectory, mol_widget=mol_widget)
            self.settings.es_mep_widget.update_canvas(item.spline,
                                                      item.barriers,
                                                      new_widget.changed_frame,
                                                      item.invert_direction,
                                                      not item.barrierless_type)

        elif isinstance(item, Compound):
            # update widget
            # get PES info
            centroid = db.Structure(item.db_representation.get_centroid(), self._structure_collection)
            charge = centroid.get_charge()
            mult = centroid.get_multiplicity()
            # get atoms
            centroid_atoms = centroid.get_atoms()
            trajectory.elements = centroid_atoms.elements
            trajectory.push_back(centroid_atoms.positions)
            self.settings.es_mep_widget.clear_canvas()
            new_widget = MoleculeVideo(parent=self, trajectory=trajectory, mol_widget=mol_widget)

        # Construct widget
        self.reset_item_colors()
        self.set_brush_pen_for_connected_items(item.db_representation.id().string())
        self.settings.update_molecule_widget(new_widget)
        # # # Highlight outline of focused item
        self.focused_item.setPen(self.hover_pen)

        # # # Insert mol_widget at correct position of layout
        self.settings.p_layout.insertWidget(mol_index, new_widget)

        if charge is not None and mult is not None:
            new_widget.setToolTip(f"Charge {charge}; Multiplicity {mult}")

    def hover_enter_function(self, _, item: QGraphicsItem) -> None:
        item_db_id_h = item.db_representation.id().string()
        if isinstance(item, Compound):
            if item_db_id_h in self.compounds:
                for same_item in self.compounds[item_db_id_h]:
                    self.set_hover(same_item)
        else:
            if item_db_id_h in self.reactions:
                for same_item in self.reactions[item_db_id_h]:
                    self.set_hover(same_item)

    def hover_leave_function(self, _, item: QGraphicsItem) -> None:
        item_db_id = item.db_representation.id().string()

        if isinstance(item, Compound):
            if item_db_id in self.compounds:
                for same_item in self.compounds[item_db_id]:
                    self.reset_hover(same_item)
        else:
            if item_db_id in self.reactions:
                for same_item in self.reactions[item_db_id]:
                    self.reset_hover(same_item)

    def menu_function(self, event, item: QGraphicsItem) -> None:
        menu = QMenu()
        popup_action = menu.addAction('Expand in Pop-Up')
        copyid_action = menu.addAction('Copy ID')
        if isinstance(item, Compound):
            copy_masm_graph_action = menu.addAction('Copy Molassembler Graph')
            move_action = menu.addAction('Move to Molecular Viewer')
            activate_action = menu.addAction('Activate compound for exploration')
            deactivate_action = menu.addAction('Deactivate compound for exploration')
        if isinstance(item, Reaction):
            has_ts = bool(item.spline is not None)
            if has_ts:
                move_ts_action = menu.addAction('Move TS to Molecular Viewer')
        _ = menu.addAction('Close Menu')
        action = menu.exec_(event.screenPos())  # type: ignore
        if action == popup_action:
            self.__expand_in_new_window(item)
        elif action == copyid_action:
            current_id = self.__get_id_of_current_item(item)
            copy_text_to_clipboard(current_id)
        elif isinstance(item, Compound):
            if action == copy_masm_graph_action:
                current_graph = self.__get_masm_graph_of_current_compound(item)
                copy_text_to_clipboard(current_graph)
            elif action == move_action:
                self.__move_structure_to_main_viewer(item)
            elif action == activate_action:
                item.db_representation.enable_exploration()
            elif action == deactivate_action:
                item.db_representation.disable_exploration()
        elif isinstance(item, Reaction):
            if has_ts:
                if action == move_ts_action:
                    self.__move_structure_to_main_viewer(item)

    def __expand_in_new_window(self, item: QGraphicsItem):
        if isinstance(item, Compound):
            from scine_heron.database.expand_widget import ExpandCompound
            selected_compound_dict = json.loads(item.db_representation.json())
            self.__expanded_views.append(ExpandCompound(None, self.db_manager, selected_compound_dict))
            self.__expanded_views[-1].setWindowTitle("Unique Conformers in Aggregate")
            self.__expanded_views[-1].show()
        elif isinstance(item, Reaction):
            from scine_heron.database.expand_widget import ExpandReaction
            selected_reaction_dict = json.loads(item.db_representation.json())
            self.__expanded_views.append(ExpandReaction(None, self.db_manager, selected_reaction_dict))
            self.__expanded_views[-1].setWindowTitle("Elementary steps in Reaction")
            self.__expanded_views[-1].show()

    def __move_structure_to_main_viewer(self, item: QGraphicsItem):
        from scine_heron import get_core_tab
        tab = get_core_tab('molecule_viewer')
        if tab is not None:
            if isinstance(item, Compound):
                db_compound = item.db_representation
                tmp_structure = db.Structure(db_compound.get_centroid(), self._structure_collection)
                tab.update_molecule(atoms=tmp_structure.get_atoms())
            elif isinstance(item, Reaction):
                db_es = db.ElementaryStep(item.assigned_es_id, self._elementary_step_collection)
                if db_es.has_transition_state():
                    tmp_structure = db.Structure(db_es.get_transition_state(), self._structure_collection)
                    tab.update_molecule(atoms=tmp_structure.get_atoms())

    def __get_id_of_current_item(self, item: QGraphicsItem):
        if isinstance(item, Compound) or isinstance(item, Reaction):
            item_id = item.db_representation.id().string()
            return item_id

    def __get_masm_graph_of_current_compound(self, item: QGraphicsItem) -> str:
        if isinstance(item, Compound):
            structures = self.db_manager.get_collection("structures")
            s = db.Structure(item.db_representation.get_centroid())
            s.link(structures)
            if not s.has_graph("masm_cbor_graph"):
                return "No graph available"
            masm_graph = s.get_graph("masm_cbor_graph")
            return masm_graph
        else:
            return "No graph available"

    def _get_elementary_step_of_reaction_from_db(self,
                                                 reaction: db.Reaction,
                                                 energy_model: db.Model,
                                                 structure_model: Union[None, db.Model] = None
                                                 ) -> Tuple[Union[db.ElementaryStep, None],
                                                            Union[Tuple[float, float], Tuple[None, None]]]:
        # NOTE: Only electronic energies for now
        elementary_step = None
        elementary_step = get_elementary_step_with_min_ts_energy(
            reaction,
            "electronic_energy",
            energy_model,
            self._elementary_step_collection,
            self._structure_collection,
            self._property_collection,
            max_barrier=np.inf,
            min_barrier=- np.inf,
            structure_model=structure_model
        )
        e_type = "electronic_energy"

        # Only query for barrier if es is not None
        if elementary_step is not None:
            barriers = get_barriers_for_elementary_step_by_type(elementary_step, e_type, energy_model,
                                                                self._structure_collection, self._property_collection)
        else:
            barriers = (None, None)

        return elementary_step, barriers



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ChemotonClassSearcher class.
"""

import re
import os
import ast
import _ast
import importlib
from collections import UserDict
from inspect import isabstract, getmembers
from typing import List, Tuple, Dict, Type, Optional

import scine_chemoton


class ChemotonClassSearcher(UserDict):
    """
    Automatically gather a list of all possible classes from scine_chemoton that are a subclass of the given type
    """

    def __init__(self, class_to_search: Type, black_list: Optional[List[Type]] = None,
                 avoid_exact_match: bool = False) -> None:
        """
        Search for all subclasses of given class in scine_chemoton and save them in this object.
        Key is a class name and value is a class type.
        """
        super().__init__()
        self.search_type: Type = class_to_search
        self.module_to_class: Dict[str, List[str]] = dict()

        for path, class_name in ChemotonClassSearcher.__get_all_classes_names_from_chemoton():
            module = importlib.import_module(path)
            module_name = module.__name__.replace("scine_chemoton.", "")
            module_name_split = module_name.split(".")
            # cut away first submodule
            if len(module_name_split) > 1:
                module_name = ".".join(module_name_split[1:])
            loaded_class = getattr(module, class_name)

            if avoid_exact_match and loaded_class == class_to_search:
                continue

            if issubclass(loaded_class, class_to_search):
                if isabstract(loaded_class):
                    continue
                if black_list is not None:
                    if any(issubclass(loaded_class, b) for b in black_list):
                        continue

                human_readable_name = self.__human_readable_class_names(class_name)
                self.data[human_readable_name] = loaded_class

                if module_name not in self.module_to_class:
                    self.module_to_class[module_name] = []
                self.module_to_class[module_name].append(human_readable_name)

    @staticmethod
    def __human_readable_class_names(class_name: str) -> str:
        # Split a string at uppercase letters
        words = [s for s in re.split("([A-Z][^A-Z]*)", class_name) if s]

        # convert all words except the first to lowercase
        for i in range(1, len(words)):
            words[i] = words[i].capitalize()

        # join with space
        return " ".join(words)

    @staticmethod
    def __get_all_classes_names_from_chemoton() -> List[Tuple[str, str]]:
        """
        Search for all classes in scine_chemoton module.
        Returns a list of Tuples of module path and class name.
        """
        classes = []
        for path in ChemotonClassSearcher.__get_files_names_from_chemoton():
            with open(path) as mf:
                tree = ast.parse(mf.read())

            # parse module path
            module_path = path.split("scine_chemoton")[-1]
            module_path = module_path.replace(".py", "")
            module_path = module_path.replace("/__init__", "")
            module_path = "scine_chemoton" + module_path.replace("/", ".")

            # parse classes names
            module_classes = [_ for _ in tree.body if isinstance(_, _ast.ClassDef)]
            classes.extend([(module_path, c.name) for c in module_classes])
        return classes

    @staticmethod
    def __get_files_names_from_chemoton() -> List[str]:
        """
        Return all .py files in scine_chemoton module.
        """
        location = None
        try:
            location = scine_chemoton.__file__  # can also be None
        except AttributeError:
            pass
        if location is None:
            members = getmembers(scine_chemoton)
            for m in members:
                if "__file__" in m and m[1] is not None:
                    location = m[1]
                    break
                if "__path__" in m and m[1] is not None:
                    for possible_location in m[1]:
                        if os.path.exists(possible_location):
                            location = possible_location
                            break
            if location is None:
                raise RuntimeError("The location of your installed SCINE Chemoton could not be determined")
        path = os.path.dirname(location)

        py_files = []

        for root, _, files in os.walk(path):
            for f in files:
                if f[-3:] == ".py":
                    py_files.append(os.path.join(root, f))

        return py_files



__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from enum import Enum


class Interfaces(Enum):

    OpenMolcas = "OpenMolcas"
    Serenity = "Serenity"


class BasicOptions:
    def __init__(self):
        self.interface = Interfaces.OpenMolcas
        self.basis_set: str = "cc-pvdz"
        self.charge: int = 0
        self.spin_multiplicity: int = 0
        self.number_of_roots: int = 1
        self.error_message: str = ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
import numpy as np
from datetime import timedelta
from json import dumps
from typing import Optional, List

from scine_database.queries import finished_calculations, unstarted_calculations

from scine_heron.multithread import Worker
from scine_heron.utilities import (
    color_axis,
    color_figure,
    get_font,
    timedelta_string
)

from PySide2.QtCore import QObject, QThreadPool
from PySide2.QtWidgets import QPushButton, QVBoxLayout, QDialog, QLabel

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure


class RuntimeHistogramDialog(QDialog):
    class Chart(FigureCanvasQTAgg):
        def __init__(self, db_manager, width=5, height=4) -> None:
            self.fig = Figure(figsize=(width, height))
            self.ax1 = self.fig.add_subplot(2, 1, 1)
            self.ax2 = self.fig.add_subplot(2, 1, 2)
            color_figure(self.fig)
            self.font = get_font()
            color_axis(self.ax2)
            color_axis(self.ax1)
            super(RuntimeHistogramDialog.Chart, self).__init__(self.fig)
            self.db_manager = db_manager
            self.ax1.cla()
            self.ax2.cla()
            self.fig.tight_layout()
            self.draw()
            self.__currently_updating = False
            self.walltime: List[float] = []
            self.cpuh: List[float] = []

        def update_complete(self):
            self.__draw_plot()
            self.__currently_updating = False

        def update_data(self):
            if not self.__currently_updating:
                self.__currently_updating = True
                worker = Worker(self.__update_data)
                worker.signals.finished.connect(self.update_complete)
                pool = QThreadPool.globalInstance()
                pool.start(worker)

        def __update_data(self, *args, **kwargs) -> None:
            calculations = self.db_manager.get_collection("calculations")
            selection = finished_calculations()
            # Loop over all results
            self.walltime = []
            self.cpuh = []
            for calculation in calculations.query_calculations(dumps(selection)):
                if calculation.has_runtime():
                    cores = calculation.get_job().cores
                    runtime = calculation.get_runtime()
                    self.walltime.append(runtime)
                    self.cpuh.append(cores * runtime)

            if not self.walltime:
                self.walltime = [1.0]
                self.cpuh = [1.0]

        def __draw_plot(self):
            min_val = np.floor(np.log10(np.min(self.walltime)))
            max_val = np.ceil(np.log10(np.max(self.walltime)))
            self.ax1.cla()
            self.ax1.hist(
                self.walltime,
                bins=10 ** np.linspace(min_val, max_val, 50),
                log=False,
                color="lightblue",
            )
            self.ax1.set_title("Walltime", self.font)
            self.ax1.set_xlabel("Walltime in s")
            self.ax1.set_ylabel("Count")
            self.ax1.set_xscale("log")

            min_val = np.floor(np.log10(np.min(self.cpuh)))
            max_val = np.ceil(np.log10(np.max(self.cpuh)))
            self.ax2.cla()
            self.ax2.hist(
                self.cpuh,
                bins=10 ** np.linspace(min_val, max_val, 50),
                log=False,
                color="lightblue",
            )
            self.ax2.set_title("CPU Time", self.font)
            self.ax2.set_xlabel("CPU Time in s")
            self.ax2.set_ylabel("Count")
            self.ax2.set_xscale("log")
            self.fig.tight_layout()
            self._add_reference_time_labels()
            self.draw()

        def _add_reference_time_labels(self):
            color = "C1"
            special_x = [60.0, 3600.0, 86400.0, 604800.0]
            labels = ['minute', 'hour', 'day', 'week']
            for ax in [self.ax1, self.ax2]:
                y_pos = np.mean(ax.get_ylim())
                for sx, label in zip(special_x, labels):
                    ax.axvline(x=sx, color=color)
                    ax.text(sx, y_pos, label,
                            fontsize=10, color=color,
                            rotation=90, rotation_mode='anchor')

    class InfoText(QLabel):
        def __init__(self, db_manager=None) -> None:
            super().__init__()
            self.db_manager = db_manager
            self.setText("Click on update to retrieve runtime information")

        def update_data(self):
            self.setText("Querying database...")
            self.repaint()  # required to update mid-function
            calculations = self.db_manager.get_collection("calculations")
            selection = finished_calculations()
            n_calcs = calculations.count(dumps(selection))
            if n_calcs == 0:
                self.setText("No finished calculations so far")
                return
            avg_wall_time = sum(
                c.runtime for c in calculations.query_calculations(dumps(selection)) if c.runtime is not None
            )
            avg_wall_time = timedelta(seconds=(avg_wall_time / n_calcs))
            text = (
                f"Total walltime so far: {timedelta_string(avg_wall_time * n_calcs)} \n"
                + f"Average walltime per job: {timedelta_string(avg_wall_time)} \n"
            )
            n_pending = calculations.count(dumps({"status": "pending"}))
            if n_pending > 0:
                avg_cores_for_pending = sum(
                    c.get_job().cores
                    for c in calculations.query_calculations(
                        dumps({"status": "pending"})
                    )
                )
                avg_cores_for_pending /= n_pending
                n_todo = calculations.count(dumps(unstarted_calculations()))
                walltime_left = (
                    (n_todo + n_pending) * avg_wall_time * avg_cores_for_pending / n_pending
                )
                workers = "puffins" if n_pending > 1 else "puffin"
                text += f"Estimated walltime left with currently {n_pending} {workers}: " + \
                        f"{timedelta_string(walltime_left)}"
                text += "\n(This estimate may be inaccurate if you have a database with varying models)"
            self.setText(text)

    def __init__(
        self,
        parent: Optional[QObject] = None,
        db_manager=None,
        window_title: str = "Runtime Histogram",
    ) -> None:
        super(RuntimeHistogramDialog, self).__init__(parent)
        self.setWindowTitle(window_title)

        # Create layout and add widgets
        layout = QVBoxLayout()
        self.chart = self.Chart(db_manager)
        layout.addWidget(self.chart)
        self.info_text = self.InfoText(db_manager)
        layout.addWidget(self.info_text)

        self.button_update = QPushButton("Update")
        layout.addWidget(self.button_update)
        self.button_update.clicked.connect(self.update_data)  # pylint: disable=no-member

        self.button_close = QPushButton("Close")
        layout.addWidget(self.button_close)
        self.button_close.clicked.connect(self.close)  # pylint: disable=no-member

        # Set dialog layout
        self.setLayout(layout)
        self.exec_()

    def update_data(self):
        self.info_text.update_data()
        self.chart.update_data()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides a number of vtkAlgorithms and their wrappers
that can be used in MoleculeDepthView classes.
"""

from typing import List, Tuple, Any
from vtk import (
    vtkMolecule,
    vtkArrayData,
    vtkTrivialProducer,
    vtkArray,
    VTK_FLOAT,
    vtkActor,
    vtkInformationVector,
    vtkAlgorithmOutput,
)
import numpy as np

from vtk.util.vtkAlgorithm import VTKPythonAlgorithmBase, vtkAlgorithm
from vtkmodules.vtkCommonDataModel import vtkPolyData


class DepthProjectionAlgorithm(VTKPythonAlgorithmBase):  # type: ignore[misc]
    """
    Takes a molecule and projects every atom to the line of sight
    of a camera.
    The center of the view is the haptic pointer.
    The atoms in the visible are selected
    based on the distance from the haptic pointer.
    The z-movement of the atoms is not uniform,
    but it is magnified close to the haptic pointer.
    """

    def __init__(self, filter_radius: float, zscale: float) -> None:
        super().__init__(
            nInputPorts=3,
            nOutputPorts=1,
            inputType=[
                "vtkPolyData",  # The position of the haptic pointer
                "vtkMolecule",  # The molecule
                "vtkArrayData",  # The camera information - depth vector
            ],
            outputType="vtkMolecule",
        )
        self.filter_radius = filter_radius
        self.zscale = zscale

    def FillInputPortInformation(self, port: int, info: Any) -> int:
        """Sets the required input type to InputType."""
        info.Set(vtkAlgorithm.INPUT_REQUIRED_DATA_TYPE(), self.InputType[port])
        return 1

    def _core_algorithm(
        self,
        haptic_position_data: vtkPolyData,
        in_molecule: vtkMolecule,
        camera_data: vtkArrayData,
        out_molecule: vtkMolecule,
    ) -> None:
        """
        An easier-to-test method that uses typed data.
        """
        # CAMERA_ARRAYS: order must match (see other CAMERA_ARRAYS label)
        depth_vector = camera_data.GetArray(0)

        # Preparing data
        haptic_position = np.array([*haptic_position_data.GetPoints().GetPoint(0)])
        depth_vector = np.array([depth_vector.GetValue(i) for i in range(3)])

        atom_positions = np.array(
            [
                in_molecule.GetAtom(i).GetPosition()
                for i in range(in_molecule.GetNumberOfAtoms())
            ]
        )

        if in_molecule.GetNumberOfAtoms() > 0:
            depths_absolute = atom_positions @ depth_vector
            depths = depths_absolute - haptic_position @ depth_vector

            position_wrt_haptic = atom_positions - haptic_position
            dist_from_haptic_sq = (position_wrt_haptic ** 2).sum(axis=1)
            plot_condition = dist_from_haptic_sq < self.filter_radius ** 2

            depths_remapped = self.depthremap(depths)
            for i in range(in_molecule.GetNumberOfAtoms()):
                if plot_condition[i]:
                    new_atom = out_molecule.AppendAtom()

                    atom = in_molecule.GetAtom(i)
                    new_atom.SetAtomicNumber(atom.GetAtomicNumber())
                    new_atom.SetPosition((0, 0, depths_remapped[i]))

    def RequestData(
        self,
        _: Any,
        in_info: List[vtkInformationVector],
        out_info: vtkInformationVector,
    ) -> int:
        """
        Performs the transformation
        """
        haptic_position = vtkPolyData.GetData(in_info[0])
        in_molecule = vtkMolecule.GetData(in_info[1])

        camera_data = vtkArrayData.GetData(in_info[2])
        out_molecule = vtkMolecule.GetData(out_info)

        self._core_algorithm(haptic_position, in_molecule, camera_data, out_molecule)

        # return True (as an integer) to signal success
        return 1

    def depthremap(self, z: np.ndarray) -> np.ndarray:
        return self.zscale * (2.0 / (1 + np.exp(-2 * z / self.zscale)) - 1.0)


class DepthProjection:
    def __init__(self, filter_range: float, zscale: float) -> None:
        """
        haptic pointer -----------------------------,
        molecule -> producer ------------------------> MoleculeDepthProjectionAlgorithm -> output
        cameradata -> (focus_position, -> producer /
                       depth_vector)
        """
        # Haptic pointer data is hooked directly to the algorithm.

        # Molecule
        self.__producer = vtkTrivialProducer()

        # Camera Data
        self.__depth_vector = vtkArray.CreateArray(vtkArray.DENSE, VTK_FLOAT)
        self.__depth_vector.Resize(3)

        self.__camera_data = vtkArrayData()
        self.__camera_data_producer = vtkTrivialProducer()

        self.__depth_projector = DepthProjectionAlgorithm(filter_range, zscale)
        # connecting all together
        # molecule not yet!

        # camera data
        # CAMERA_ARRAYS: order must match (see other CAMERA_ARRAYS label)
        self.__camera_data.AddArray(self.__depth_vector)

        self.__camera_data_producer.SetOutput(self.__camera_data)
        # Algorithm
        self.__depth_projector.SetInputConnection(1, self.__producer.GetOutputPort())
        self.__depth_projector.SetInputConnection(
            2, self.__camera_data_producer.GetOutputPort()
        )

    @property
    def output(self) -> vtkActor:
        return self.__depth_projector.GetOutputPort()

    def set_molecule(self, molecule: vtkMolecule) -> None:
        """
        Setting a vtkMolecule as input,
        using the internale vtkTrivialProducer.
        """
        self.__producer.SetOutput(molecule)
        self.__depth_projector.SetInputConnection(1, self.__producer.GetOutputPort())

    def set_molecule_input(self, output_port: vtkAlgorithmOutput) -> None:
        """
        Setting directly the input connection
        for the molecule input port
        of the depth projection algorithm.
        """
        self.__depth_projector.SetInputConnection(1, output_port)

    def set_haptic_pointer_data(self, output_port: vtkAlgorithmOutput) -> None:
        """
        Setting directly the input connection
        for the haptic pointer data input port
        of the depth projection algorithm.
        """
        self.__depth_projector.SetInputConnection(0, output_port)

    def set_camera_data(self, depth_vector: Tuple[float, float, float],) -> None:

        for i in range(3):
            self.__depth_vector.SetValue(i, depth_vector[i])
        self.__camera_data.Modified()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import Any, List, Union, Tuple
import re
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg

import scine_database as db

from scine_chemoton.utilities.get_molecular_formula import get_molecular_formula_of_aggregate

from scine_heron.database.energy_diagram import EnergyDiagram
from scine_heron.toolbar.io_toolbar import HeronToolBar
from scine_heron.io.file_browser_popup import get_save_file_name

from PySide2.QtWidgets import (
    QWidget,
    QCheckBox,
    QGraphicsView,
    QVBoxLayout,
    QHBoxLayout,
    QGraphicsScene,
    QScrollArea,
    QStyle,
)
from PySide2.QtGui import (
    QGuiApplication
)
from PySide2.QtCore import Qt


class CustomQScrollArea(QScrollArea):
    def wheelEvent(self, event):
        modifiers = QGuiApplication.keyboardModifiers()
        if self.widget().mol_widget.underMouse() and modifiers == Qt.ControlModifier:
            self.widget().mol_widget.wheelEvent(event)
        elif self.widget().mol_widget.underMouse():
            super(CustomQScrollArea, self).wheelEvent(event)
        else:
            super(CustomQScrollArea, self).wheelEvent(event)


class PathLevelWidget(QWidget):
    def __init__(self, parent: QWidget, db_manager: db.Manager,
                 levels: List[float], barrierless: List[bool], type_list: List[Any],
                 path_info: Tuple[int, float, List[Any]], rxn_equation: str, es_method_info: str) -> None:
        self.window_width = 1000
        self.window_height = 800
        self.scene_object = QGraphicsScene(0, 0, self.window_width, self.window_height)
        super(PathLevelWidget, self).__init__(parent=parent)
        self.db_manager = db_manager

        layout_main = QVBoxLayout()
        layout = QVBoxLayout()
        self.path_level_diagram = PathLevelDiagramWidget(self, self.db_manager, path_info[2],
                                                         levels, barrierless, type_list, rxn_equation, es_method_info)
        layout.addWidget(self.path_level_diagram)

        self.path_level_settings = PathLevelDiagramSettings(self, self.path_level_diagram)
        layout.addWidget(self.path_level_settings)
        # Add 1st Row
        layout_main.addLayout(layout)

        self.view = QGraphicsView(self.scene_object)
        self.view.setLayout(layout_main)
        self.view.setWindowTitle(
            "Path #" + str(path_info[0]) + " - Length " + str(path_info[1]))
        self.view.show()


class PathLevelDiagramWidget(QWidget):
    def __init__(self, parent: QWidget, db_manager: db.Manager, path: List[Any],
                 levels: List[float], barrierless: List[bool], type_list: List[Any], rxn_equation: str,
                 es_method_info: str) -> None:
        self.window_width = 1000
        self.window_height = 800
        self.scene_object = QGraphicsScene(0, 0, self.window_width, self.window_height)
        super(PathLevelDiagramWidget, self).__init__(parent=parent)
        self.db_manager = db_manager

        # db-info of selected compound
        self._compound_collection = self.db_manager.get_collection("compounds")
        self._flask_collection = self.db_manager.get_collection("flasks")
        self._structure_collection = self.db_manager.get_collection("structures")
        # # # private variables
        self.path = path
        self.levels = levels
        self.barrierless_list = barrierless
        self.type_list = type_list
        self.overall_rxn_equation = rxn_equation
        self.es_method_info = es_method_info

        layout_main = QVBoxLayout()

        layout = QHBoxLayout()

        self.fig = Figure(figsize=(self.window_width * 0.9, self.window_height * 0.9))
        self.ax1 = self.fig.add_subplot(1, 1, 1)
        self.canvas = FigureCanvasQTAgg(self.fig)
        self.canvas.setMinimumHeight(250)
        self.canvas.setMinimumWidth(400)

        layout.addWidget(self.canvas)

        layout_main.addLayout(layout)

        self.setLayout(layout_main)

        self.generate_energy_diagram()
        self.set_x_label()
        self._format_plot(self.ax1)

    @staticmethod
    def _format_plot(ax):

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.yaxis.set_ticks_position('left')
        ax.xaxis.set_ticks_position('bottom')
        # Ax lw normally 1.0
        ax.spines['bottom'].set_color('black')
        ax.spines['bottom'].set_linewidth(1.5)
        ax.spines['left'].set_color('black')
        ax.spines['left'].set_linewidth(1.5)
        ax.tick_params(direction='out', length=5, width=1.5, colors='black')

    def set_x_label(self):
        # Format reaction as well
        new_eq = ""
        first = True
        # Loop over sides and reactants to have easier string to regex
        if self.overall_rxn_equation != " = ":  # If no reaction is given, do not try to format it
            for side in self.overall_rxn_equation.split(" = "):
                new_side = ""
                for reactant in side.split(" + "):
                    tmp_factor = reactant.split(" ")[0]
                    tmp_reactant = re.sub(r'(\d+)', r'$_{\1}$', reactant.split(" ")[1].split("(")[0])
                    tmp_charge_multi = "(" + reactant.split("(")[1]
                    new_side += tmp_factor + " " + tmp_reactant + tmp_charge_multi + " + "
                if first:
                    new_side = new_side[:-3] + " = "
                    first = False
                    new_eq += new_side
                else:
                    new_eq += new_side[:-3]
        else:
            new_eq = self.overall_rxn_equation

        self.ax1.set_xlabel(f"{new_eq}\n{self.es_method_info}")

    def _construct_molecular_formula(self, aggregate_id: str, aggregate_type: db.CompoundOrFlask):

        aggregate_str = get_molecular_formula_of_aggregate(
            db.ID(aggregate_id), aggregate_type,
            self._compound_collection, self._flask_collection, self._structure_collection)
        split_aggregate = aggregate_str.split("[")
        if len(split_aggregate) == 1:
            # make subscript and remove (c: x, m: y)
            m_formula = re.sub(r'(\d+)', r'$_{\1}$', split_aggregate[0].split("(")[0].strip())
        else:
            # make subscript
            compounds_string = split_aggregate[1][:-1].replace(" | ", "\n+ ")
            m_formula = re.sub(r'(\d+)', r'$_{\1}$', compounds_string)

        return m_formula

    def generate_energy_diagram(self, print_bottom_formulas: bool = True):
        diagram = EnergyDiagram(9 / 16)  # self.window_width * 0.9/self.window_height * 0.9)
        diagram.dimension = 100
        diagram.top_text_fontsize = 'small'
        diagram.bottom_text_fontsize = 'small'
        digit_round = 1
        diagram.round_energies_at_digit = digit_round
        current_index = 0
        level_index = 0
        # Color scheme
        # TODO adapt to dark mode
        # TODO: Maybe as general setting
        default_line_color = '#85929e'
        barrierless_color = '#c2c9cf'
        ts_level_color = '#995151'
        if print_bottom_formulas:
            starting_state_formula = self._construct_molecular_formula(self.path[0], self.type_list[0])
        else:
            starting_state_formula = ""
        type_list_counter = 1
        # Ground Level at the beginning
        diagram.add_level(self.levels[level_index], bottom_text=starting_state_formula,
                          color=default_line_color, linewidth=4.5,
                          )
        # Skipping does not work yet
        for barrierless in self.barrierless_list:
            current_energy = self.levels[level_index]
            if not barrierless:
                level_index += 1
                # Derive relative change
                rel_energy_change = self.levels[level_index] - current_energy
                tmp_text = "+" if rel_energy_change > 0 else ""
                tmp_text += str(round(rel_energy_change, diagram.round_energies_at_digit))
                # Add TS level
                # Different Color for TS
                diagram.add_level(self.levels[level_index], '', top_text=tmp_text,
                                  color=ts_level_color, linewidth=4.5, linestyle='-')
                current_energy = self.levels[level_index]
                # Link up
                diagram.add_link(current_index, current_index + 1, linewidth=2.0, color=default_line_color)
                current_index += 1
                # Link Down
                diagram.add_link(current_index, current_index + 1, linewidth=2.0, color=default_line_color)
                current_index += 1
                level_index += 1
            else:
                # Jump to next ground state
                level_index += 2
                # Link ground states
                diagram.add_link(current_index, current_index + 1, linewidth=2.0, color=barrierless_color)
                current_index += 1
            # Derive relative change
            # NOTE: maybe make absolute energy optional as well
            rel_energy_change = self.levels[level_index] - current_energy
            tmp_text = "+" if rel_energy_change > 0 else ""
            tmp_text += str(round(rel_energy_change, diagram.round_energies_at_digit))
            tmp_bottom_level_info = ""
            if print_bottom_formulas:
                # # # Absolute energy to bottom
                tmp_bottom_level_info = str(round(self.levels[level_index], digit_round))
                tmp_bottom_level_info += "\n"
                tmp_bottom_level_info += self._construct_molecular_formula(self.path[level_index],
                                                                           self.type_list[type_list_counter])
            type_list_counter += 1  # Counter requires to be increased
            # Add next ground state
            diagram.add_level(self.levels[level_index], bottom_text=tmp_bottom_level_info, top_text=tmp_text,
                              color=default_line_color, linewidth=4.5, linestyle='-')
        # Plot diagram
        range_yaxis = abs(max(self.levels) - min(self.levels))
        custom_ylim = (min(self.levels) - 0.35 * range_yaxis,
                       max(self.levels) + 0.1 * range_yaxis)
        diagram.plot(ax=self.ax1, ylimits=custom_ylim)


class PathLevelDiagramSettings(QWidget):
    def __init__(self, parent: QWidget,
                 diagram_widget: Any
                 ) -> None:
        super(PathLevelDiagramSettings, self).__init__(parent=parent)
        self.p_layout = QHBoxLayout()
        # Elements of widget relative to width of ReactionProfile Canvas
        self._widget_width = 300
        # Fixed width for widget
        self.setMinimumWidth(self._widget_width)
        self.setMaximumWidth(self._widget_width)

        self.diagram_widget = diagram_widget
        self.status_widget: Union[None, QWidget] = None

        # Export svg diagram
        self.toolbar = HeronToolBar(parent=self)
        self.toolbar.shortened_add_action(
            QStyle.SP_DialogSaveButton, "Save Diagram", "Ctrl+S", self._save_svg
        )
        self.p_layout.addWidget(self.toolbar)

        self.show_formulas_of_level = QCheckBox("Show molecular formula")
        self.show_formulas_of_level.setChecked(True)
        self.show_formulas_of_level.stateChanged.connect(self._trigger_diagram_update)  # pylint: disable=no-member
        self.p_layout.addWidget(self.show_formulas_of_level)

        self.setLayout(self.p_layout)
        self.show()

    def _trigger_diagram_update(self):
        self.diagram_widget.ax1.cla()
        if self.show_formulas_of_level.isChecked():
            self.diagram_widget.generate_energy_diagram(print_bottom_formulas=True)
        else:
            self.diagram_widget.generate_energy_diagram(print_bottom_formulas=False)
        self.diagram_widget.set_x_label()
        self.diagram_widget.canvas.draw()

    def _save_svg(self):
        filename = get_save_file_name(self, "energy_diagram", ["svg"])
        self.diagram_widget.canvas.draw()
        self.diagram_widget.fig.savefig(filename, bbox_inches='tight', dpi=300, transparent=True)



__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import TYPE_CHECKING, Any, Optional

from PySide2.QtCore import QObject

from scine_heron.autocas.basic_options import BasicOptions

if TYPE_CHECKING:
    Signal = Any
else:
    pass

from PySide2.QtCore import Signal


class OptionsStatusManager(QObject):
    interface_changed_signal = Signal()
    status_bar_update_signal = Signal(str)

    def __init__(self, basic_options: BasicOptions, parent: Optional[QObject] = None):
        super().__init__(parent)
        self.__basic_options = basic_options

    @property
    def basic_options(self) -> BasicOptions:
        """
        Returns the settings.
        """
        return self.__basic_options

    @property
    def error_message(self) -> str:
        """
        Returns the error_message.
        """
        return self.__basic_options.error_message

    @error_message.setter
    def error_message(self, value: str) -> None:
        """
        Sets the contained value. Notifies on change.
        """
        if value == self.__basic_options.error_message or value == "":
            return

        self.__basic_options.error_message = value
        self.status_bar_update_signal.emit(value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""


from typing import Optional, List, Tuple

from PySide2.QtWidgets import (
    QVBoxLayout,
    QWidget,
)

from scine_utilities import AtomCollection
from scine_heron.molecule.molecule_widget import MoleculeWidget


class StructureWithReactiveSites(QWidget):
    """
    A wrapper widget around a molecular viewer that highlights reactive sites.

    Notes
    -----
    Structure modifications are disabled.
    """

    color: Tuple[float, float, float] = (78., 128., 207.)
    """
    The color of the reactive site highlighting.
    """

    def __init__(self, parent: Optional[QWidget], atoms: Optional[AtomCollection] = None,
                 reactive_indices: Optional[List[int]] = None) -> None:
        """
        Constructor the widget, structures and reactive sites can be modified later.

        Parameters
        ----------
        parent : Optional[QWidget]
            The parent widget
        atoms : Optional[AtomCollection], optional
            The structure, by default None
        reactive_indices : Optional[List[int]], optional
            The reactive indices that should be hightlighted, by default None
        """
        super().__init__(parent)
        self._mol_widget = MoleculeWidget(parent=self, atoms=atoms, disable_modification=True)
        if atoms is not None and reactive_indices is not None:
            self.update_structure(atoms, reactive_indices)
        layout = QVBoxLayout()
        layout.addWidget(self._mol_widget)
        self.setLayout(layout)
        self.setMinimumHeight(250)

    def clear(self):
        """
        Removes the molecule
        """
        self._mol_widget.update_molecule()
        self._mol_widget.reset_camera()

    def reset_camera(self):
        """
        Resets the camera in the molecule widget
        """
        self._mol_widget.reset_camera()

    def update_structure(self, atoms: AtomCollection, reactive_indices: List[int]) -> None:
        """
        Plug in a new structure and reactive sites.
        The current structure and indices are removed.

        Parameters
        ----------
        atoms : AtomCollection
            The new structure
        reactive_indices : List[int]
            The new reactive indices
        """
        self._mol_widget.update_molecule(atoms=atoms)
        assert len(self.color) == 3
        color = (self.color[0] / 255.0, self.color[1] / 255.0, self.color[2] / 255.0)
        self._mol_widget.set_selection(reactive_indices, color=color)

    def get_structure(self) -> AtomCollection:
        """
        Get the current structure as an AtomCollection.

        Returns
        -------
        AtomCollection
            The current structure
        """
        return self._mol_widget.get_atom_collection()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from PySide2.QtWidgets import QStyle, QToolBar, QWidget

from scine_heron.autocas.autocas_settings import AutocasSettings
from scine_heron.autocas.signal_handler import SignalHandler


class ToolbarWidget(QToolBar):
    # pylint: disable-next=W0613
    def __init__(
        self, parent: QWidget, signal_handler: SignalHandler, settings: AutocasSettings
    ):
        super().__init__()
        self.signal_handler = signal_handler
        self.autocas_settings = settings

        # self.open_molecule_file = self.addAction(
        #     self.style().standardIcon(QStyle.SP_TitleBarMenuButton),
        #     self.tr(self.__action_padding("Open xyz file")),  # type: ignore
        # )
        # self.open_entanglement_view.triggered.connect(
        #    self.signal_handler.open_entanglement_widget_signal
        # )

        self.start_autocas_calculation = self.addAction(
            self.style().standardIcon(QStyle.SP_MediaPlay),
            self.tr(self.__action_padding("Start Autocas calculation")),  # type: ignore
        )
        self.open_entanglement_view.triggered.connect(
            self.signal_handler.open_entanglement_widget_signal
        )

        self.open_settings_view = self.addAction(
            self.style().standardIcon(QStyle.SP_FileDialogContentsView),
            self.tr(self.__action_padding("Open Settings")),  # type: ignore
        )
        # self.open_entanglement_view.triggered.connect(
        #    self.signal_handler.open_entanglement_widget_signal
        # )

        self.open_filebrowser_view = self.addAction(
            self.style().standardIcon(QStyle.SP_DirHomeIcon),
            self.tr(self.__action_padding("Open File Browser")),  # type: ignore
        )
        # self.open_entanglement_view.triggered.connect(
        #    self.signal_handler.open_entanglement_widget_signal
        # )

        self.open_output_view = self.addAction(
            self.style().standardIcon(QStyle.SP_TrashIcon),
            self.tr(self.__action_padding("Open Output")),  # type: ignore
        )
        # self.open_entanglement_view.triggered.connect(
        #    self.signal_handler.open_entanglement_widget_signal
        # )
        self.open_file_tree = self.addAction(
            self.style().standardIcon(QStyle.SP_TrashIcon),
            self.tr(self.__action_padding("Open File Tree")),  # type: ignore
        )
        self.open_file_tree.triggered.connect(
            self.signal_handler.toggle_file_tree
        )

    @staticmethod
    def __action_padding(annotation: str) -> str:
        return '<p style="color:black !important;">{}</p>'.format(annotation)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional, List

from PySide2.QtCore import QObject
from PySide2.QtWidgets import (
    QDialog,
    QLabel,
    QCheckBox,
)
from scine_heron.containers.without_wheel_event import NoWheelDoubleSpinBox
from scine_chemoton.utilities.model_combinations import ModelCombination
from scine_chemoton.utilities.place_holder_model import construct_place_holder_model
from scine_chemoton.utilities.uncertainties import (
    UncertaintyEstimator, ConstantUncertainty, ModelCombinationBasedUncertaintyEstimator, ZeroUncertainty
)

from scine_heron.chemoton.uncertainty_tuple import UncertaintyTuple
from scine_heron.chemoton.model_combination_builder import ModelCombinationBuilder
from scine_heron.containers.layouts import VerticalLayout, HorizontalLayout
from scine_heron.settings.class_options_widget import DictOptionWidget


class UncertaintyEstimatorBuilder(QDialog):
    def __init__(self, option: UncertaintyEstimator, parent: Optional[QObject] = None):
        """
        Build the options object for the uncertainty estimator.
        """
        super().__init__(parent)
        self._uncertainty_list: List[UncertaintyTuple] = [UncertaintyTuple(
            ModelCombination(construct_place_holder_model()), ConstantUncertainty((0.0, 0.0)),
            ConstantUncertainty((0.0, 0.0)))
        ]
        use_default_zero = True
        if isinstance(option, ModelCombinationBasedUncertaintyEstimator):
            uncertainties_for_models = option.get_model_uncertainties()
            if uncertainties_for_models:
                if not all(isinstance(t[1], ConstantUncertainty) and isinstance(t[2], ConstantUncertainty)
                           for t in uncertainties_for_models):
                    raise ValueError("Only constant uncertainties are supported in Heron at the moment.")
                self._uncertainty_list = [UncertaintyTuple(t[0], t[1], t[2]) for t in uncertainties_for_models]
                use_default_zero = False
        self._disclaimer_widget = QLabel("Only constant uncertainties are supported in Heron at the moment.\n"
                                         "Use a Python script directly for more sophisticated uncertainty\n"
                                         "quantification.", parent=parent)
        self._use_default_zero_uncertainty = QCheckBox("Assume zero uncertainties for all models", parent=parent)
        self._use_default_zero_uncertainty.toggled.connect(self.__box_check_function)  # pylint: disable=no-member

        self._label = "Uncertainties"
        options = {self._label: self._uncertainty_list}
        self.dict_option_widget = DictOptionWidget(options, parent=parent)
        self._uncertainty_list_widget, self._getter_function\
            = self.dict_option_widget.construct_widget_based_on_type(self._uncertainty_list, self._label)

        layout = VerticalLayout([
            self._disclaimer_widget,
            self._use_default_zero_uncertainty,
            self._uncertainty_list_widget
        ])
        self.setLayout(layout)
        self._use_default_zero_uncertainty.setChecked(use_default_zero)

    def __box_check_function(self) -> None:
        self._uncertainty_list_widget.setVisible(not self._use_default_zero_uncertainty.isChecked())

    def get_uncertainty_estimator(self):
        if self._use_default_zero_uncertainty.isChecked():
            return ZeroUncertainty()
        uncertainty_tuple_list = [t.get_tuple() for t in self._getter_function()]
        return ModelCombinationBasedUncertaintyEstimator(uncertainty_tuple_list)


class UncertaintyBuilder(QDialog):
    def __init__(self, option: UncertaintyTuple, parent: Optional[QObject] = None):
        """
        Build the widget for an element of a list of uncertainties. This is required by the uncertainty estimator.
        """
        super().__init__(parent)
        assert isinstance(option, UncertaintyTuple)
        self._model_combination = option.model_combination
        self._free_energy_uncertainty = option.free_energy_uncertainty
        self._activation_uncertainty = option.activation_energy_uncertainty
        self._model_combination_label = QLabel("Model combination")
        self._model_combination_widget = ModelCombinationBuilder(self._model_combination)
        self._free_energy_uncertainty_label = QLabel("Free Energy Uncertainty (in J/mol)")
        if (not isinstance(self._free_energy_uncertainty, ConstantUncertainty)
                or not isinstance(self._activation_uncertainty, ConstantUncertainty)):
            raise ValueError("Only constant uncertainties are supported in Heron at the moment.")
        self._free_energy_uncertainty_widget = ConstantUncertaintyBuilder(self._free_energy_uncertainty, parent)
        self._activation_uncertainty_label = QLabel("Activation Energy Uncertainty (in J/mol)")
        self._activation_uncertainty_widget = ConstantUncertaintyBuilder(self._activation_uncertainty, parent)

        layout = VerticalLayout([
            self._model_combination_label,
            self._model_combination_widget,
            self._free_energy_uncertainty_label,
            self._free_energy_uncertainty_widget,
            self._activation_uncertainty_label,
            self._activation_uncertainty_widget
        ])
        self.setLayout(layout)

    def get_uncertainty_tuple(self) -> UncertaintyTuple:
        return UncertaintyTuple(self._model_combination_widget.get_model_combination(),
                                self._free_energy_uncertainty_widget.get_uncertainty(),
                                self._activation_uncertainty_widget.get_uncertainty())


class ConstantUncertaintyBuilder(QDialog):
    def __init__(self, option: ConstantUncertainty, parent: Optional[QObject] = None):
        """
        Build a widget for a constant uncertainty object.
        """
        super().__init__(parent)
        self._label = "uncertainty_bounds"

        lower, upper = option.get_uncertainty_bounds()
        maximum = 999999
        self._lower_edit = NoWheelDoubleSpinBox(parent=parent)
        self._lower_edit.setMinimum(0.0)
        self._lower_edit.setMaximum(maximum)
        self._lower_edit.setSingleStep(0.01)
        self._lower_edit.setDecimals(2)
        self._lower_edit.setValue(lower)

        self._upper_edit = NoWheelDoubleSpinBox(parent=parent)
        self._upper_edit.setMinimum(0.0)
        self._upper_edit.setMaximum(maximum)
        self._upper_edit.setSingleStep(0.01)
        self._upper_edit.setDecimals(2)
        self._upper_edit.setValue(upper)

        self._layout = HorizontalLayout([
            self._lower_edit,
            self._upper_edit
        ])
        self.setLayout(self._layout)

    def get_uncertainty(self) -> ConstantUncertainty:
        return ConstantUncertainty((self._lower_edit.value(), self._upper_edit.value()))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from PySide2.QtCore import Qt, QTimer, QPoint, QThreadPool, SignalInstance, QDate  # QTimeLine
from PySide2.QtGui import QPainterPath, QGuiApplication
from PySide2.QtWidgets import (
    QWidget,
    QPushButton,
    QLabel,
    QDateTimeEdit,
    QLineEdit,
    QGraphicsPathItem,
    QVBoxLayout,
    QHBoxLayout,
    QSplitter,
    QCheckBox,
    QScrollArea
)
from datetime import datetime
from scine_heron import find_main_window
from scine_heron.settings.class_options_widget import ModelOptionsWidget
from scine_heron.database.reaction_compound_view import ReactionAndCompoundView, ReactionAndCompoundViewSettings
from scine_heron.database.graphics_items import Compound, Reaction
from scine_heron.containers.layouts import VerticalLayout, HorizontalLayout
from scine_heron.containers.buttons import TextPushButton
from scine_heron.io.file_browser_popup import get_load_file_name
from scine_heron.multithread import Worker
from scine_chemoton.gears.pathfinder import Pathfinder as pf

from scine_database.concentration_query_functions import (
    query_concentration,
    query_reaction_flux
)
from scine_database.energy_query_functions import (
    get_energy_change,
    get_barriers_for_elementary_step_by_type,
)
import scine_database as db
import copy
import networkx as nx
import numpy as np
from json import dumps
from typing import Dict, Optional, Any, List, Union, Tuple, TYPE_CHECKING


if TYPE_CHECKING:
    Slot = Any
else:
    from PySide2.QtCore import Slot


class AdvancedSettingsWidget(QWidget):
    def __init__(self, network, db_manager, parent=None) -> None:
        super(AdvancedSettingsWidget, self).__init__(parent)
        self._max_barrier = 2650.5
        self._min_barrier = - 1000.0
        self._always_show_barrierless = True
        self._model: Optional[db.Model] = None
        self._model_button = QPushButton("Set Model")
        self._model_button.clicked.connect(self._call_model_dialog)  # pylint: disable=no-member
        self._enforce_structure_model = False
        self._scale_with_concentrations = False
        self._min_flux = 1e-5
        self._network = network
        self._db_manager = db_manager

        self.__layout = QVBoxLayout()

        self._set_up_barrier_widgets(self.__layout)

        self.__layout.addWidget(self._model_button)

        self.enforce_structure_model_cbox = QCheckBox("Enforce Structure Model", parent=self)
        self.enforce_structure_model_cbox.setChecked(self._enforce_structure_model)
        self.__layout.addWidget(self.enforce_structure_model_cbox)

        self._set_up_concentration_widgets(self.__layout)

        self.setLayout(self.__layout)

    def get_max_barrier(self) -> float:
        return float(self.current_max_barrier_text.text())

    def get_min_barrier(self) -> float:
        return float(self.current_min_barrier_text.text())

    def always_show_barrierless(self) -> bool:
        return self.always_show_barrierless_reactions_cbox.isChecked()

    def get_min_flux(self) -> float:
        return float(self.current_min_flux_text.text())

    def get_model(self) -> db.Model:
        if self._model is None:
            self._query_for_general_model()
        if self._model is None:
            self._call_model_dialog()
        assert self._model is not None
        return self._model

    def scale_with_concentrations(self):
        return self.concentration_scaling_cbox.isChecked()

    def enforce_structure_model(self):
        return self.enforce_structure_model_cbox.isChecked()

    def _call_model_dialog(self):
        dialog = ModelOptionsWidget(self, self._model)
        # Note: Ideally, the option widget should still be movable
        dialog.exec_()
        # Overwrite model from dialog after exec if the initial is None
        if self._model is None:
            self._model = copy.deepcopy(dialog._options)

    def _set_up_barrier_widgets(self, layout):
        qhbox_labels = QHBoxLayout()
        qhbox_texts = QHBoxLayout()
        self.current_max_barrier_label = QLabel(self)
        self.current_max_barrier_label.resize(280, 40)
        self.current_max_barrier_label.setText("Max. Barrier Height [kJ/mol]")
        qhbox_labels.addWidget(self.current_max_barrier_label)
        self.current_max_barrier_text = QLineEdit(self)
        self.current_max_barrier_text.resize(280, 40)
        self.current_max_barrier_text.setText(str(self._max_barrier))
        qhbox_texts.addWidget(self.current_max_barrier_text)

        self.current_min_barrier_label = QLabel(self)
        self.current_min_barrier_label.resize(280, 40)
        self.current_min_barrier_label.setText("Min. Barrier Height [kJ/mol]")
        qhbox_labels.addWidget(self.current_min_barrier_label)
        self.current_min_barrier_text = QLineEdit(self)
        self.current_min_barrier_text.resize(280, 40)
        self.current_min_barrier_text.setText(str(self._min_barrier))
        qhbox_texts.addWidget(self.current_min_barrier_text)

        layout.addLayout(qhbox_labels)
        layout.addLayout(qhbox_texts)
        self.always_show_barrierless_reactions_cbox = QCheckBox("Always show barrier-less reactions", parent=self)
        self.always_show_barrierless_reactions_cbox.setChecked(self._always_show_barrierless)
        layout.addWidget(self.always_show_barrierless_reactions_cbox)

    def _set_up_concentration_widgets(self, layout):
        self.concentration_scaling_cbox = QCheckBox("Truncate by Concentration")
        self.concentration_scaling_cbox.setChecked(self._scale_with_concentrations)
        layout.addWidget(self.concentration_scaling_cbox)
        self.current_min_flux_label = QLabel(self)
        self.current_min_flux_label.resize(280, 40)
        self.current_min_flux_label.setText("Min. Concentration Flux")
        layout.addWidget(self.current_min_flux_label)
        self.current_min_flux_text = QLineEdit(self)
        self.current_min_flux_text.resize(280, 40)
        self.current_min_flux_text.setText(str(self._min_flux))
        layout.addWidget(self.current_min_flux_text)

    def _query_for_general_model(self):
        selection = {
            "$and": [
                {"status": {"$eq": "complete"}},
                {"results.elementary_steps.0": {"$exists": True}},
            ]
        }
        calculation = (
            self._db_manager.get_collection("calculations").get_one_calculation(dumps(selection))
        )
        if calculation is None:
            selection = {"label": {"$in": ["user_optimized", "user_surface_optimized"]}}
            structure = self._db_manager.get_collection("structures").get_one_structure(dumps(selection))
            if structure is None:
                return
            model = structure.model
        else:
            model = calculation.model
        # # # Generalize initial model
        for model_attribute in ["electronic_temperature", "pressure", "spin_mode", "temperature", "version"]:
            setattr(model, model_attribute, "any")
        self._model = model

    def update_settings(self):
        """
        Read the model definition and maximum barrier from the input boxes. If "None" is given in the method family
        input box. Get some model from the database.
        """
        self._enforce_structure_model = self.enforce_structure_model()
        # Maximum/Minimum barrier and flux.
        self._max_barrier = self.get_max_barrier()
        self._min_barrier = self.get_min_barrier()
        self._always_show_barrierless = self.always_show_barrierless()
        self._min_flux = self.get_min_flux()
        self._scale_with_concentrations = self.scale_with_concentrations()


class CRNetwork(ReactionAndCompoundView):
    def __init__(self, parent: QWidget, db_manager: db.Manager) -> None:
        super().__init__(parent=parent)
        self.setMouseTracking(True)
        self.setInteractive(True)
        self.setMinimumWidth(100)
        self.setMinimumHeight(100)

        self.structure_of_compound = None

        self._graph: Union[nx.DiGraph, None] = None
        self._graph_model = db.Model("any", "any", "any")
        self._graph_with_structure_model = False
        self._currently_plotting = False
        self.new_network = False
        # Cache of subgraphs with graph and positions
        # NOTE: Don't know how memory intensive
        self.subgraph_cache: Dict[str, Dict[str, Any]] = dict()

        # Add all data
        self.db_manager: db.Manager = db_manager
        self.centroid_item: Optional[Compound] = None
        self.current_centroid_id: str = ""
        self.__history: List[str] = []
        self.__current_history_position: int = -1

        # db-info of selected compound
        self._compound_collection = self.db_manager.get_collection("compounds")
        self._reaction_collection = self.db_manager.get_collection("reactions")
        self._flask_collection = self.db_manager.get_collection("flasks")
        self._elementary_step_collection = self.db_manager.get_collection("elementary_steps")
        self._property_collection = self.db_manager.get_collection("properties")
        self._structure_collection = self.db_manager.get_collection("structures")

    @staticmethod
    def _print_progress(signal):

        main_window = find_main_window()
        if main_window is not None:
            status_bar = main_window.get_status_bar()
            if signal[0]:
                status_bar.clear_message()
                status_bar.update_status(signal[1], timer=None)
            # When signal is false, error status
            else:
                status_bar.clear_message()
                status_bar.update_error_status(signal[1], timer=5 * 1000)

    @staticmethod
    def trigger_thread_function(func, info_func, *args, **kwargs):
        worker = Worker(func, *args, **kwargs)
        worker.signals.running.connect(info_func)

        pool = QThreadPool.globalInstance()
        pool.start(worker)

    def focus_function(self, _, item: Union[Compound, Reaction]) -> None:
        # Store current connected items
        self.subgraph_cache[self.current_centroid_id]["focused_connected_items"] = self.focused_connected_items
        self.subgraph_cache[self.current_centroid_id]["focused_connected_lines"] = self.focused_connected_lines
        id_string = item.db_representation.id().string()
        assert self.settings
        cr_settings = self.settings
        cr_settings.update_current_centroid_text(id_string)
        cr_settings.advanced_settings_widget.update_settings()
        self.trigger_thread_function(
            self.update_network,
            self._print_progress,
            cr_settings.advanced_settings_widget.get_model(),
            cr_settings.advanced_settings_widget.enforce_structure_model(),
            trigger_plot=True,
            requested_centroid=id_string,
            triggered_from_focus_click=True,
        )

    def redo_move(self):
        if (self.__current_history_position + 2) > len(self.__history) or self.__current_history_position < 0:
            return
        self.__current_history_position += 1
        centroid_id = self.__history[self.__current_history_position]
        self.update_network(self._graph_model, self._graph_with_structure_model,
                            trigger_plot=True,
                            requested_centroid=centroid_id,
                            track_update=False)
        # Update text field
        self.settings.update_current_centroid_text(
            str(self.current_centroid_id)
        )

    def undo_move(self):
        if self.__current_history_position > len(self.__history) or self.__current_history_position < 1:
            return
        self.__current_history_position -= 1
        centroid_id = self.__history[self.__current_history_position]
        self.update_network(self._graph_model, self._graph_with_structure_model,
                            trigger_plot=True,
                            requested_centroid=centroid_id,
                            track_update=False)
        # Update text field
        self.settings.update_current_centroid_text(
            str(self.current_centroid_id)
        )

    def load_graph(self):
        pathfinder = pf(self.db_manager)
        graph_filenname = get_load_file_name(self, "graph", ['json'])
        if graph_filenname is None:
            return
        # Load new graph
        pathfinder.load_graph(graph_filenname)
        self._graph = copy.deepcopy(pathfinder.graph_handler.graph)
        self.settings.advanced_settings_widget._call_model_dialog()
        self._graph_model = self.settings.advanced_settings_widget.get_model()
        # NOTE: Assume structure model for loaded graph
        self.settings.advanced_settings_widget._enforce_structure_model = True
        self.settings.advanced_settings_widget.enforce_structure_model_cbox.setChecked(True)
        self._graph_with_structure_model = True

    def reset_graph(self):
        self._graph = None
        self._graph_model = db.Model("any", "any", "any")

    def reset_subgraph_cache(self):
        self.subgraph_cache = {}

    def remove_all_items(self):
        # # # Remove old items from current scene
        items = self.scene_object.items()
        for item in items:
            self.scene_object.removeItem(item)

    def plot_network(self):
        if self._currently_plotting:
            return
        self._currently_plotting = True
        self.remove_all_items()
        # # # Add new items
        for k in self.line_items.keys():
            self.scene_object.addItem(self.line_items[k])
        for c in self.compounds.keys():
            self.scene_object.addItem(self.compounds[c][0])
        for r in self.reactions.keys():
            self.scene_object.addItem(self.reactions[r][0])
        self.__center_view()

        self._currently_plotting = False

    def update_network(
        self,
        model: db.Model,
        enforce_structure_model: bool,
        progress_callback: Optional[SignalInstance] = None,
        trigger_plot=False,
        requested_centroid: Optional[str] = None,
        track_update: bool = True,
        triggered_from_focus_click: bool = False
    ) -> None:
        if progress_callback is not None:
            progress_callback.emit((True, "Status:" + 4 * " " + "Loading Network"))
        # If graph not present, build basic graph
        if self._graph is None or \
           model != self._graph_model or \
           enforce_structure_model != self._graph_with_structure_model:
            # NOTE: possible to introduce subgraph cache per model
            self.reset_subgraph_cache()
            # NOTE: Add loading screen
            self.__build_graph()
            assert self._graph
            if len(self._graph.nodes()) == 0:
                assert progress_callback
                progress_callback.emit((False, 'Status:' + 4 * " " + "No graph available!"))
                return

            assert self._graph

        if requested_centroid is None:
            tmp_compound = self._compound_collection.get_one_compound(dumps({}))
            if tmp_compound:
                requested_centroid = tmp_compound.id().string()
            else:
                assert progress_callback
                progress_callback.emit((False, 'Status:' + 4 * " " + "No compounds available!"))
                return

        # Check that requested centroid is in graph
        if requested_centroid not in self._graph.nodes():
            assert progress_callback
            progress_callback.emit((False, 'Status:' + 4 * " " + "Centroid not available!"))
            return
        else:
            self.current_centroid_id = requested_centroid

        # Reset storage
        self.compounds = {}
        self.reactions = {}
        self.line_items = {}

        if track_update:
            if len(self.__history) == 0:
                self.__current_history_position += 1
                self.__history = self.__history[0:(self.__current_history_position)]
                self.__history.append(self.current_centroid_id)
            elif self.__history[-1] != self.current_centroid_id:
                self.__current_history_position += 1
                self.__history = self.__history[0:self.__current_history_position]
                self.__history.append(self.current_centroid_id)

        # # # Start building subgraph
        centroid_sub_graph, reaction_nodes, aggregate_nodes = self.__get_subgraph_of_centroid(
            self.current_centroid_id)
        # NOTE: Start building subgraph of all aggregate nodes as subthread (process?),
        #       just list of nodes required as input

        dist_value = 0.2
        max_dimensions = (420, 420)
        if len(centroid_sub_graph.nodes) > 400:
            dist_value = 0.1
            max_dimensions = (int(1000 * 1.5), 1000)

        # # # Retrieve positions
        if "positions" not in self.subgraph_cache[self.current_centroid_id].keys():

            positions = self.__get_node_positions_for_subgraph(self.current_centroid_id, centroid_sub_graph,
                                                               reaction_nodes, aggregate_nodes, dist_value)
            scaled_positions = self.__scale_positions(positions, max_dimensions)

            # Store scaled positions in cache
            self.subgraph_cache[self.current_centroid_id]["positions"] = scaled_positions
            # Init Cache for compounds and reactions
            self.subgraph_cache[self.current_centroid_id]["compounds"] = dict()
            self.subgraph_cache[self.current_centroid_id]["reactions"] = dict()
            self.subgraph_cache[self.current_centroid_id]["lines"] = dict()
            self.subgraph_cache[self.current_centroid_id]["focused_connected_items"] = list()
            self.subgraph_cache[self.current_centroid_id]["focused_connected_lines"] = list()
        else:
            # Load positions from cache:
            scaled_positions = copy.deepcopy(self.subgraph_cache[self.current_centroid_id]["positions"])
            # Load connected items from previous scene
            self.focused_connected_items = copy.deepcopy(self.subgraph_cache[self.current_centroid_id]
                                                         ["focused_connected_items"])
            self.focused_connected_lines = copy.deepcopy(self.subgraph_cache[self.current_centroid_id]
                                                         ["focused_connected_lines"])
            if len(self.focused_connected_items) > 0:
                self.view_highlighted = True

        # # # Build centroid item
        aggregate, a_type = self.__get_aggregate_and_type_from_graph(requested_centroid)
        self.centroid_item, _ = self.__build_compound_item(0, 0, aggregate, a_type, allow_focus=False)

        # # # Build all items
        self.__build_subgraph_items(centroid_sub_graph, scaled_positions, reaction_nodes)
        # # # Reset colors of items if coming from a focus function call
        if triggered_from_focus_click:
            self.reset_item_colors()

        # Update with current centroid id, if empty
        if self.settings.current_id_text.text() == "":
            self.settings.update_current_centroid_text(str(self.current_centroid_id))

        if trigger_plot:
            self.new_network = True

        if progress_callback is not None:
            progress_callback.emit((True, 'Status:' + 4 * " " + "Network Ready"))

    def __build_compound_item(self, x: int, y: int, db_compound: Union[db.Compound, db.Flask],
                              a_type: db.CompoundOrFlask, allow_focus=True) -> Tuple[Compound, str]:
        cid_string = db_compound.id().string()
        # Look up in cache first
        if self.current_centroid_id in self.subgraph_cache.keys() and\
           cid_string in self.subgraph_cache[self.current_centroid_id]['compounds'].keys():
            old_compound_item = self.subgraph_cache[self.current_centroid_id]['compounds'][cid_string]
            # # # Copy stored compound item - necessary for scaling in this thread
            compound_item = Compound(old_compound_item.x_coord, old_compound_item.y_coord,
                                     pen=old_compound_item.get_pen(), brush=old_compound_item.get_brush(),
                                     concentration=old_compound_item.concentration)
            compound_item.set_current_brush(old_compound_item.get_current_brush())
            compound_item.set_current_pen(old_compound_item.get_current_pen())
            compound_item.db_representation = old_compound_item.db_representation
            compound_item.set_created(old_compound_item.get_created())
            compound_item.final_concentration = old_compound_item.final_concentration
            compound_item.concentration_flux = old_compound_item.concentration_flux
            # delete old_compound_item
            del old_compound_item
        else:
            centroid_structure = db.Structure(db_compound.get_centroid(), self._structure_collection)
            c_max = query_concentration("max_concentration", centroid_structure, self._property_collection)
            compound_item = Compound(x, y, pen=self.compound_pen, brush=self.get_aggregate_brush(a_type),
                                     concentration=c_max)
            compound_item.db_representation = db_compound
            compound_item.set_created(db_compound.created())

            compound_item.final_concentration = query_concentration("final_concentration",
                                                                    centroid_structure, self._property_collection)
            compound_item.concentration_flux = query_concentration("concentration_flux",
                                                                   centroid_structure, self._property_collection)

        # Allow to focus is different from graph travel
        self.__bind_functions_to_object(compound_item, allow_focus)

        # Decide upon scaling
        if self.settings.advanced_settings_widget.scale_with_concentrations():
            compound_item.allow_scaling = True
            compound_item.add_concentration_tooltip()
        else:
            compound_item.allow_scaling = False
        # Update size of compound item
        compound_item.update_size()
        # Add to cache
        self.subgraph_cache[self.current_centroid_id]['compounds'][cid_string] = compound_item
        self.replace_in_compound_list(compound_item)

        return compound_item, cid_string

    def __build_reaction_item(self, x: int, y: int, reaction: db.Reaction, side: int):
        assert self._graph
        assert self.settings
        plot_item = True
        # Look up in cache first
        if reaction.id().string() in self.subgraph_cache[self.current_centroid_id]['reactions'].keys():
            old_reaction_item = self.subgraph_cache[self.current_centroid_id]['reactions'][reaction.id().string()]
            # # # Copy stored reaction item - necessary for scaling in this thread
            reaction_item = Reaction(old_reaction_item.x_coord, old_reaction_item.y_coord,
                                     pen=old_reaction_item.get_pen(), brush=old_reaction_item.get_brush(),
                                     flux=old_reaction_item.get_flux(),
                                     invert_direction=old_reaction_item.invert_direction)
            reaction_item.set_current_brush(old_reaction_item.get_current_brush())
            reaction_item.set_current_pen(old_reaction_item.get_current_pen())
            # Copy all attributes
            reaction_item.barriers = old_reaction_item.barriers
            reaction_item.db_representation = old_reaction_item.db_representation
            reaction_item.set_created(old_reaction_item.get_created())
            reaction_item.assigned_es_id = old_reaction_item.assigned_es_id
            reaction_item.energy_difference = old_reaction_item.energy_difference
            reaction_item.spline = old_reaction_item.spline
            reaction_item.barrierless_type = old_reaction_item.barrierless_type
            # delete old_reaction_item
            del old_reaction_item
        else:
            # Get elementary step from graph or database
            es_from_graph = None
            node_id = reaction.id().string() + ";" + str(side) + ";"
            if "elementary_step_id" in self._graph.nodes(data=True)[node_id]:
                es_id = db.ID(
                    self._graph.nodes(data=True)[node_id]["elementary_step_id"])
                es_from_graph = db.ElementaryStep(es_id, self._elementary_step_collection)

            if es_from_graph is None:
                if self._graph_with_structure_model:
                    tmp_structure_model = self._graph_model
                else:
                    tmp_structure_model = None
                es, barriers = self._get_elementary_step_of_reaction_from_db(
                    reaction,
                    self._graph_model,
                    tmp_structure_model)
            else:
                es = es_from_graph
                barriers = get_barriers_for_elementary_step_by_type(es, "electronic_energy", self._graph_model,
                                                                    self._structure_collection,
                                                                    self._property_collection)

            if es is None:
                return None
            # Add reaction flux to reaction item
            # NOTE: Absolute flux, i guess
            flux = query_reaction_flux("_reaction_edge_flux", reaction,
                                       self._compound_collection, self._flask_collection,
                                       self._structure_collection, self._property_collection)
            # Creating reaction item
            reaction_item = Reaction(
                x, y, flux=flux,
                pen=self.reaction_pen,
                brush=self.get_reaction_brush(es.get_type()),
                # invert direction if side is 1
                invert_direction=bool(side)
            )
            reaction_item.set_barriers(barriers)

            if es.get_type() == db.ElementaryStepType.BARRIERLESS:
                reaction_item.barrierless_type = True

            reaction_item.db_representation = reaction
            reaction_item.set_created(reaction.created())
            reaction_item.assigned_es_id = es.id()
            struct = db.Structure(es.get_reactants()[0][0], self._structure_collection)
            # NOTE: Energy difference might be redundant
            reaction_item.energy_difference = get_energy_change(
                es, "electronic_energy", struct.get_model(), self._structure_collection, self._property_collection)

            if es.has_spline():
                reaction_item.spline = es.get_spline()

        self.__bind_functions_to_object(reaction_item)
        # Decide upon scaling
        if self.settings.advanced_settings_widget.scale_with_concentrations():
            reaction_item.allow_scaling = True
        else:
            reaction_item.allow_scaling = False

        # Decide, if plotted or not
        plot_item = False
        if reaction_item.barriers is not None:
            plot_item = self.__barrier_within_plot_window(reaction_item.barriers, side) and \
                self.__created_after_time_limit(reaction_item.get_created()) and \
                self.__reaction_flux_above_threshold(reaction_item.get_flux())
        # Check, if barrierless should be shown
        if reaction_item.barrierless_type and not\
           self.settings.advanced_settings_widget.always_show_barrierless():
            plot_item = False
        # Add to cache
        self.subgraph_cache[self.current_centroid_id]['reactions'][reaction.id().string()] = reaction_item
        if plot_item:
            self.add_to_reaction_list(reaction_item)
            return reaction_item
        else:
            return None

    def __build_edge(self, start: QPoint, end: QPoint, line_id: str):
        # Look edge up in cache first
        if line_id in self.subgraph_cache[self.current_centroid_id]['lines'].keys():
            edge_item = self.subgraph_cache[self.current_centroid_id]['lines'][line_id]
        else:
            path = QPainterPath(start)
            path.lineTo(end)
            # ID must be unique for scene!
            edge_item = QGraphicsPathItem(path)
            edge_item.setPen(self.path_pen)
            self.subgraph_cache[self.current_centroid_id]['lines'][line_id] = edge_item
        return edge_item

    def __barrier_within_plot_window(self, barriers: Union[Tuple[float, float], Tuple[None, None]],
                                     side: int):
        acceptable_barrier = True
        # Safety Check to avoid None comparison
        if not any(barrier is None for barrier in barriers) and self.settings is not None:
            if self.settings.advanced_settings_widget.get_max_barrier() < barriers[side] or \
               self.settings.advanced_settings_widget.get_min_barrier() > barriers[side]:
                acceptable_barrier = False

        return acceptable_barrier

    def __created_after_time_limit(self, created: Union[datetime, None]) -> bool:
        if created is None or created > datetime.fromtimestamp(self.settings.time_edit.dateTime().toSecsSinceEpoch()):
            return True
        else:
            return False

    def __reaction_flux_above_threshold(self, flux: Union[float, None]) -> bool:
        if self.settings.advanced_settings_widget.scale_with_concentrations() and \
           flux <= self.settings.advanced_settings_widget.get_min_flux():
            return False
        else:
            return True

    def __build_graph(self):
        # Init basic finder
        pathfinder = pf(self.db_manager)
        pathfinder.options.barrierless_weight = 1.0
        # Set model for graph from widget
        self._graph_model = copy.deepcopy(self.settings.advanced_settings_widget.get_model())
        pathfinder.options.model = self._graph_model
        pathfinder.options.graph_handler = "basic"

        pathfinder.options.use_structure_model = False
        if self.settings.advanced_settings_widget.enforce_structure_model():
            pathfinder.options.use_structure_model = True
            pathfinder.options.structure_model = self._graph_model
            self._graph_with_structure_model = True
        else:
            self._graph_with_structure_model = False
        # Build graph
        pathfinder.build_graph()
        self._graph = copy.deepcopy(pathfinder.graph_handler.graph)

    def __get_subgraph_of_centroid(self, centroid_id: str) -> Tuple[nx.DiGraph, List[str], List[str]]:
        assert self._graph
        reaction_nodes = []
        aggregate_nodes = []
        # Attempt reset
        centroid_subgraph = None
        undirected_graph = self._graph.to_undirected()
        if centroid_id not in self.subgraph_cache.keys():
            # # # Reactions outgoing from current centroid only!
            for centroid_edge in self._graph.out_edges(centroid_id):
                reaction_nodes.append(centroid_edge[1])
                for reaction_edge in undirected_graph.edges(centroid_edge[1]):
                    if reaction_edge[1] != centroid_id:
                        aggregate_nodes.append(reaction_edge[1])
            # Extract from subgraph from graph
            centroid_subgraph = self._graph.subgraph([centroid_id] + reaction_nodes + aggregate_nodes).copy()
            # Reset weights to one
            # NOTE: If not basic, with barrier info, one could attempt incorporating this information
            for edge in centroid_subgraph.edges:
                centroid_subgraph.edges[edge]['weight'] = 1.0  # old reset
            # Store subgraph in cache
            self.subgraph_cache[centroid_id] = {"graph": centroid_subgraph}
        # # # End building Subgraph
        else:
            centroid_subgraph = copy.deepcopy(self.subgraph_cache[centroid_id]["graph"])
            # Sort to list of nodes
            for node in centroid_subgraph.nodes():
                if ";" in node:
                    reaction_nodes.append(node)
                else:
                    if node != centroid_id:
                        aggregate_nodes.append(node)

        return centroid_subgraph, reaction_nodes, aggregate_nodes

    @staticmethod
    def __get_node_positions_for_subgraph(centroid_id: str, subgraph: nx.DiGraph,
                                          reaction_node_ids: List[str], aggregate_node_ids: List[str],
                                          min_node_distance=0.2) -> Dict[str, Tuple[float, float]]:
        # # # Start Positioning
        # Initial shell, with two rings
        # Loop over aggregates just to get all positions
        shell_position = nx.shell_layout(subgraph, [[centroid_id], reaction_node_ids, aggregate_node_ids],
                                         center=(0, 0), scale=1.0, rotate=0.0)

        # Determine subshell positioning
        for node in reaction_node_ids:
            # Products of reaction outside of rxn node
            outer_sub_shell_rxn = [r_edge[1] for r_edge in subgraph.out_edges(node) if r_edge[1] != centroid_id]
            if len(outer_sub_shell_rxn) > 1:
                outer_center = shell_position[node] + 0.3 * shell_position[node]
            else:
                outer_center = shell_position[node] + 0.25 * shell_position[node]
            outer_sub_shell_pos = nx.shell_layout(subgraph, [outer_sub_shell_rxn],
                                                  center=outer_center, scale=0.05)
            # Overwrite shell positions of outer sub shells
            for key, value in outer_sub_shell_pos.items():
                shell_position[key] = value

            # Reagents or reaction inside of rxn node
            inner_sub_shell_rxn = [r_edge[0] for r_edge in subgraph.in_edges(node) if r_edge[0] != centroid_id]
            if len(inner_sub_shell_rxn) > 1:
                inner_center = shell_position[node] - 0.3 * shell_position[node]
            else:
                inner_center = shell_position[node] - 0.25 * shell_position[node]
            inner_sub_shell_pos = nx.shell_layout(subgraph, [inner_sub_shell_rxn],
                                                  center=inner_center, scale=0.05)
            # Overwrite shell positions of inner sub shells
            for key, value in inner_sub_shell_pos.items():
                shell_position[key] = value

        # Determine position

        spring_position = nx.spring_layout(
            subgraph.to_undirected(),
            pos=shell_position,
            k=min_node_distance,  # node distance
            fixed=[centroid_id],
        )

        return copy.deepcopy(spring_position)

    @staticmethod
    def __scale_positions(positions: Dict[str, Any],
                          max_dimensions: Tuple[int, int]) -> Dict[str, Tuple[int, int]]:
        pos_matrix_list = []
        pos_key = []
        for key, value in positions.items():
            pos_key.append(key)
            pos_matrix_list.append(value)

        pos_matrix = np.asarray(pos_matrix_list)
        # Save normalization
        norm_x = np.max(np.abs(pos_matrix[:, 0])) if np.max(np.abs(pos_matrix[:, 0])) > 1e-9 else 1.0
        norm_y = np.max(np.abs(pos_matrix[:, 1])) if np.max(np.abs(pos_matrix[:, 1])) > 1e-9 else 1.0
        pos_matrix[:, 0] = pos_matrix[:, 0] / norm_x * max_dimensions[0]
        pos_matrix[:, 1] = pos_matrix[:, 1] / norm_y * max_dimensions[1]

        scaled_position = {}
        for key, value in zip(pos_key, pos_matrix):
            scaled_position[key] = value.astype(int)

        return scaled_position

    def __get_aggregate_and_type_from_graph(self,
                                            node_id: str) -> Tuple[Union[db.Compound, db.Flask], db.CompoundOrFlask]:
        assert self._graph
        aggregate: Union[db.Compound, db.Flask]
        if self._graph.nodes(data=True)[node_id]['type'] == db.CompoundOrFlask.COMPOUND.name:
            a_type = db.CompoundOrFlask.COMPOUND
            aggregate = db.Compound(db.ID(node_id), self._compound_collection)
        else:
            a_type = db.CompoundOrFlask.FLASK
            aggregate = db.Flask(db.ID(node_id), self._flask_collection)

        return aggregate, a_type

    def __build_subgraph_items(self, subgraph: nx.DiGraph,
                               positions: Dict[str, Tuple[int, int]], reaction_nodes: List[str]):
        # NOTE: parallelize this!
        for node in reaction_nodes:
            # Draw Rxn Nodes and edges
            rxn = db.Reaction(db.ID(node.split(";")[0]), self._reaction_collection)
            side = int(node.split(";")[1])
            # Filter are in building the reaction item
            rxn_item = self.__build_reaction_item(positions[node][0], positions[node][1], rxn, side)
            if rxn_item is None:
                continue
            # # # Rotate rxn item such that the rhs side points to the average outgoing position
            out_positions = np.array([np.array(positions[edge[1]]) for edge in subgraph.out_edges(node)])
            # Obtain relative vector of all outgoing positions
            rel_out_positions = np.sum(out_positions - np.array(positions[node]), axis=0)
            # Calculate angle to match the outgoing of the rxn item
            # with the average relative vector of the products
            # range from 0 to 2*pi, arctan(cross.norm, dot)
            rot_angle = np.arctan2(-rel_out_positions[1], np.dot(rel_out_positions, np.array([1, 0])))
            rot_angle += 2 * np.pi if rot_angle < 0 else 0
            rxn_item.update_angle(rot_angle)
            # # # Draw incoming edges of rxn node
            for e_count, edge in enumerate(subgraph.in_edges(node)):
                # Draw edge from aggregate to lhs of rxn
                line_id = "_".join(list(edge)) + "_" + str(e_count)
                edge_item = self.__build_edge(QPoint(positions[edge[0]][0],
                                                     positions[edge[0]][1]),
                                              rxn_item.incoming(), line_id)
                self.line_items[line_id] = edge_item

                # Check if compound is drawn already
                if edge[0] not in self.compounds.keys():
                    # Retrieve information of aggregate
                    aggregate, a_type = self.__get_aggregate_and_type_from_graph(edge[0])
                    a_position = positions[edge[0]]
                    self.__build_compound_item(a_position[0], a_position[1], aggregate, a_type)

            # # # Draw outgoing edges of rxn node
            for e_count, edge in enumerate(subgraph.out_edges(node)):
                # Draw edge from rhs of rxn to aggregate
                line_id = "_".join(list(edge)) + "_" + str(e_count)
                edge_item = self.__build_edge(rxn_item.outgoing(),
                                              QPoint(positions[edge[1]][0],
                                                     positions[edge[1]][1]),
                                              line_id)
                self.line_items[line_id] = edge_item
                # Check if compound is drawn already, else draw
                if edge[1] not in self.compounds.keys():
                    # Retrieve information of aggregate
                    aggregate, a_type = self.__get_aggregate_and_type_from_graph(edge[1])
                    a_position = positions[edge[1]]
                    self.__build_compound_item(a_position[0], a_position[1], aggregate, a_type)

    def __center_view(self):
        # self.fitInView(self.sceneRect(), Qt.KeepAspectRatio)
        self.centerOn(QPoint(0, 0))

    def __bind_functions_to_object(self, object: Any, allow_focus: bool = False) -> None:
        object.bind_mouse_press_function(self.mouse_press_function)
        if allow_focus:
            object.bind_mouse_double_click_function(self.focus_function)
        object.setAcceptHoverEvents(True)
        object.bind_hover_enter_function(self.hover_enter_function)
        object.bind_hover_leave_function(self.hover_leave_function)
        object.bind_menu_function(self.menu_function)

    def __getitem__(self, i):
        # get information of database to make subscriptable
        return f"{i}"


class CRSettings(ReactionAndCompoundViewSettings):

    def __init__(self, parent: QWidget, network: CRNetwork) -> None:
        super().__init__(parent, VerticalLayout())
        self.network = network
        self.new_network = False
        self.button_query = TextPushButton("Query Database", self._query_database_function)
        self.button_update = TextPushButton("Update View", self._update_function, shortcut="F5")
        self.button_load = TextPushButton("Load Graph", self._load_graph_function, shortcut="Ctrl+L")
        self.button_undo = TextPushButton("Undo", self.network.undo_move, shortcut="Ctrl+Z")
        self.button_redo = TextPushButton("Redo", self.network.redo_move, shortcut="Ctrl+R")

        self.time_label = QLabel(self)
        self.time_label.resize(100, 20)
        self.time_label.setFixedHeight(20)
        self.time_label.setText("Only Show Modified Reactions Since")

        self.time_edit = QDateTimeEdit(QDate())
        self.time_edit.setDisplayFormat("HH:mm dd.MM.yyyy")

        self.current_id_label = QLabel(self)
        self.current_id_label.resize(100, 20)
        self.current_id_label.setFixedHeight(20)
        self.current_id_label.setText("Current Center Aggregate")

        self.current_id_text = QLineEdit(self)
        self.current_id_text.resize(100, 40)
        self.current_id_text.setText(str(self.network.current_centroid_id))

        self.save_to_svg_button = TextPushButton("Save SVG", self.network.save_svg, shortcut="Ctrl+S")
        self.button_traveling = TextPushButton("Start Path Analysis", self.__start_path_analysis_function)

        # Timer for reploting
        self.check_new_network_timer = QTimer(self)
        self.check_new_network_timer.setInterval(50)
        self.check_new_network_timer.timeout.connect(self.__plot_new_network)  # pylint: disable=no-member
        if not self.check_new_network_timer.isActive():
            self.check_new_network_timer.start()

        self.mol_widget.setMinimumWidth(200)
        self.mol_widget.setMaximumWidth(1000)
        self.mol_widget.setMinimumHeight(300)
        self.mol_widget.setMaximumWidth(1000)

        self.es_mep_widget.setMinimumHeight(300)
        self.es_mep_widget.setMinimumWidth(200)
        self.es_mep_widget.setMaximumWidth(1000)

        self._settings_visible = False

        # set layout
        self.p_layout.addWidget(self.button_query)
        self.p_layout.addWidget(self.button_load)
        self.p_layout.addWidget(self.button_update)
        self.p_layout.addLayout(HorizontalLayout([self.button_undo, self.button_redo]))
        self.p_layout.add_widgets([
            self.time_label,
            self.time_edit,
            self.current_id_label,
            self.current_id_text
        ])
        self.p_layout.addLayout(HorizontalLayout([self.save_to_svg_button, self.button_traveling]))
        self.p_layout.add_widgets([
            self.mol_widget,
            self.es_mep_widget
        ])
        self._set_up_advanced_settings_widgets(self.p_layout)

        self._scroll_area = QScrollArea()
        scroll_widget = QWidget()
        scroll_widget.setLayout(self.p_layout)
        self._scroll_area.setWidget(scroll_widget)
        self._scroll_area.setWidgetResizable(True)
        self._scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.vscrollbar = self._scroll_area.verticalScrollBar()
        self.vscrollbar.rangeChanged.connect(self.scroll)
        self.set_advanced_settings_visible()
        layout = QVBoxLayout()
        layout.addWidget(self._scroll_area)

        self.setLayout(layout)

        self.button_deactivate_all = QPushButton("Deactivate All Aggregates")
        self.p_layout.addWidget(self.button_deactivate_all)
        self.button_deactivate_all.clicked.connect(self.__deactivate_all_aggregates)  # pylint: disable=no-member
        self.button_deactivate_all.setToolTip("Deactivates the exploration of all aggregates.\n"
                                              "Note that a running KineticsGear may activate their exploration again.")

        self.button_activate_all_user = QPushButton("Activate All USER_OPTIMIZED")
        self.p_layout.addWidget(self.button_activate_all_user)
        self.button_activate_all_user.clicked.connect(self.__activate_all_user_optimized)  # pylint: disable=no-member
        self.button_activate_all_user.setToolTip("WARNING: Activates all aggregates for exploration for which the\n"
                                                 "first structures (centroid) is labeled as USER_OPTIMIZED and NOT\n"
                                                 "all structures that have a structure with the label USER_OPTIMIZED.")
        self.show()

    @Slot(int, int)
    def scroll(self, minimum, maximum):
        value = maximum if self._settings_visible else minimum
        self.vscrollbar.setValue(value)

        self.show()

    def set_advanced_settings_visible(self) -> None:
        self._settings_visible = self.advanced_settings_cbox.isChecked()
        if self._settings_visible:
            self.advanced_settings_widget.setVisible(True)
        else:
            self.advanced_settings_widget.setVisible(False)

    def _set_up_advanced_settings_widgets(self, layout):
        self.advanced_settings_cbox = QCheckBox("ADVANCED SETTINGS")
        self.advanced_settings_cbox.setChecked(self._settings_visible)
        self.advanced_settings_cbox.toggled.connect(  # pylint: disable=no-member
            self.set_advanced_settings_visible
        )

        self.advanced_settings_widget = AdvancedSettingsWidget(
            self.network,
            self.network.db_manager
        )
        layout.addWidget(self.advanced_settings_cbox)
        layout.addWidget(self.advanced_settings_widget)

    def __deactivate_all_aggregates(self):
        compounds = self.network.db_manager.get_collection('compounds')
        flasks = self.network.db_manager.get_collection('flasks')
        for compound in compounds.iterate_all_compounds():
            compound.link(compounds)
            compound.disable_exploration()
        for flask in flasks.iterate_all_flasks():
            flask.link(flasks)
            flask.disable_exploration()

    def __activate_all_user_optimized(self):
        compounds = self.network.db_manager.get_collection('compounds')
        structures = self.network.db_manager.get_collection('structures')
        flasks = self.network.db_manager.get_collection('flasks')
        for compound in compounds.iterate_all_compounds():
            compound.link(compounds)
            centroid_id = compound.get_centroid()
            centroid = db.Structure(centroid_id, structures)
            if centroid.get_label() == db.Label.USER_OPTIMIZED:
                compound.enable_exploration()
        for flask in flasks.iterate_all_flasks():
            flask.link(flasks)
            centroid_id = flask.get_centroid()
            centroid = db.Structure(centroid_id, structures)
            if centroid.get_label() == db.Label.USER_OPTIMIZED:
                flask.enable_exploration()

    def update_current_centroid_text(self, new_text: str) -> None:
        self.current_id_text.setText(new_text)

    def _query_database_function(self) -> None:
        self.network.remove_all_items()
        self.network.reset_graph()
        self.network.reset_subgraph_cache()
        self._update_function()

    def _update_function(self) -> None:
        self.advanced_settings_widget.update_settings()
        requested_string: Union[None, str] = self.current_id_text.text()
        requested_string = requested_string if requested_string else None
        self.network.trigger_thread_function(
            self.network.update_network,
            self.network._print_progress,
            self.advanced_settings_widget.get_model(),
            self.advanced_settings_widget.enforce_structure_model(),
            trigger_plot=True,
            requested_centroid=requested_string,
        )

    def __start_path_analysis_function(self) -> None:
        from scine_heron.database.graph_traversal import GraphTravelWidget
        GraphTravelWidget(self.parent(), self.network.db_manager, self.network._graph_model)

    def _load_graph_function(self) -> None:
        self.network.remove_all_items()
        self.network.reset_graph()
        self.network.reset_subgraph_cache()
        self.network.load_graph()

    def __plot_new_network(self):
        if self.network.new_network:
            self.network.plot_network()
        self.network.new_network = False


class CustomQScrollArea(QScrollArea):
    def wheelEvent(self, event):
        modifiers = QGuiApplication.keyboardModifiers()
        if self.widget().mol_widget.underMouse() and modifiers == Qt.ControlModifier:
            self.widget().mol_widget.wheelEvent(event)
        elif self.widget().mol_widget.underMouse():
            super(CustomQScrollArea, self).wheelEvent(event)
        else:
            super(CustomQScrollArea, self).wheelEvent(event)


class CompoundReactionWidget(QWidget):
    def __init__(self, parent: QWidget, db_manager: db.Manager) -> None:
        super(CompoundReactionWidget, self).__init__(parent=parent)
        self.db_manager = db_manager

        # Create layout and add widgets
        layout = QHBoxLayout()
        self.network = CRNetwork(self, self.db_manager)
        self.settings = CRSettings(self, self.network)
        self.network.set_settings_widget(self.settings)
        self.splitter = QSplitter(self)
        self.splitter.addWidget(self.network)
        self.splitter.addWidget(self.settings)
        self.splitter.setSizes([320, 150])
        layout.addWidget(self.splitter)

        # Set dialog layout
        self.setLayout(layout)

    def center_on_aggregate(self, aggregate_id: str):
        self.settings.update_current_centroid_text(aggregate_id)
        self.settings._update_function()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from functools import partial
from os import path, mkdir, getcwd
from typing import Any, List, Optional, Dict, TYPE_CHECKING
from threading import Thread
from time import sleep

from PySide2.QtWidgets import (
    QWidget,
    QScrollArea,
    QVBoxLayout,
    QStyle,
    QSplitter,
)
from PySide2.QtCore import Qt, QObject, QThread
from PySide2.QtGui import QCloseEvent

if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal

import scine_readuct as readuct
import scine_utilities as su

from scine_heron.containers.layouts import HorizontalLayout
from scine_heron.io.text_box import yes_or_no_question
from scine_heron.io.file_browser_popup import get_load_file_name, get_save_file_name
from scine_heron.toolbar.io_toolbar import ToolBarWithSaveLoad
from scine_heron.utilities import write_error_message, clear_status_bar, write_info_message

from .create_readuct_task import CreateReaductTaskWidget
from .task_widget import TaskWidget
from .calculator_container import CalculatorContainer


class ReaductTab(QWidget):
    """
    Container Widget to handle ReaDuct systems and tasks
    """

    input_file_dir_name: str = 'input_files'

    def __init__(self, parent: Optional[QObject]) -> None:
        QWidget.__init__(self, parent)
        self._joiner: Optional[TaskJoiner] = None
        self._delete_join_threads: List[Thread] = []

        # systems
        self._calc_container = CalculatorContainer(self)

        # generate the widgets to build widgets
        self.__task_creator_widget = CreateReaductTaskWidget(self)

        # create widget for scroll area
        scroll_area_content = QWidget()
        self._scroll_area_content_layout = QVBoxLayout()
        # add toolbar
        self._toolbar = ToolBarWithSaveLoad(self._save_file, self._load_file, self)
        # play button
        self._play = self._toolbar.shortened_add_action(QStyle.SP_MediaPlay, "Start all tasks (consecutively)",
                                                        "Ctrl+M", lambda: self._start_stop_all(start=True))
        self._stop = self._toolbar.shortened_add_action(QStyle.SP_MediaPause, "Stop all tasks",
                                                        "Ctrl+R", lambda: self._start_stop_all(start=False,
                                                                                               force_join=True))
        self._play.setCheckable(True)
        self._stop.setCheckable(True)
        self.is_blocked = False
        self._scroll_area_content_layout.addWidget(self._toolbar)
        # add holder of all the created widgets
        self.created_readuct_task_widgets: List[TaskWidget] = []
        self._created_readuct_widgets_holder = QWidget()
        self.__created_readuct_widgets_layout = HorizontalLayout()
        self.__created_readuct_widgets_layout.setAlignment(Qt.AlignLeft)
        self._created_readuct_widgets_holder.setLayout(self.__created_readuct_widgets_layout)
        # finish scroll area content
        self._scroll_area_content_layout.addWidget(self._created_readuct_widgets_holder)
        scroll_area_content.setLayout(self._scroll_area_content_layout)

        # create scroll area
        self._scroll_area = QScrollArea()
        self._scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self._scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self._scroll_area.setWidget(scroll_area_content)
        self._scroll_area.setWidgetResizable(True)

        # fill total grid layout of the tab
        right_splitter = QSplitter(Qt.Orientation.Vertical)
        right_splitter.addWidget(self.__task_creator_widget)
        right_splitter.addWidget(self._scroll_area)

        total_splitter = QSplitter(Qt.Orientation.Horizontal, parent=self)
        total_splitter.addWidget(self._calc_container)
        total_splitter.addWidget(right_splitter)
        total_splitter.setSizes([320, 280])

        total_layout = QVBoxLayout()
        total_layout.addWidget(total_splitter)
        self.setLayout(total_layout)

        self._calc_container.new_system_added.connect(
            self.__task_creator_widget.add_possible_input_system
        )

    def get_systems(self) -> Dict[str, su.core.Calculator]:
        return self._calc_container.get_systems()

    def closeEvent(self, event: QCloseEvent) -> None:
        if self._runner is not None:
            self._runner.wait()
            self._joiner = TaskJoiner(self, force_join=True)
            self._joiner.start()
            self._joiner.wait()
        if self._joiner is not None:
            self._joiner.wait()
        super().closeEvent(event)

    def update_systems(self, inputs: List[str], outputs: List[str], systems: Dict[str, su.core.Calculator]) -> None:
        existing_systems = self._calc_container.get_augmented_systems()
        method_family = None
        program = None
        for inp in inputs:
            if inp not in existing_systems:
                continue
            method_family, program = existing_systems[inp][:2]
            break
        if method_family is None or program is None:
            write_error_message("Failed to update systems due to improper input information")
            return
        for out in outputs:
            if out not in systems:
                write_error_message("Failed to update systems due to improper output information")
                return
            calc = systems[out]
            self._calc_container.add_item(method_family, program, calc.structure, calc.settings.as_dict(), out)

    def add_widget(self, readuct_widget: TaskWidget) -> None:
        # connect functionalities
        if readuct_widget.button_delete is not None:
            readuct_widget.button_delete.clicked.connect(  # pylint: disable=no-member
                partial(self.delete_widget, readuct_widget)
            )
        readuct_widget.update_systems.connect(  # pylint: disable=no-member
            self.update_systems
        )

        self.created_readuct_task_widgets.append(readuct_widget)
        self.__created_readuct_widgets_layout.addWidget(readuct_widget)

    def delete_widget(self, task_widget: TaskWidget) -> int:
        idx = self.created_readuct_task_widgets.index(task_widget)
        removed_outputs = task_widget.outputs

        for widget in self.created_readuct_task_widgets[idx + 1:]:
            if any(inp in removed_outputs for inp in widget.inputs):
                write_error_message(f"You cannot delete this task, because its outputs are used by task {widget.name}!")
                return -1

        self.created_readuct_task_widgets.pop(idx)
        task_widget.stop_class_if_working()
        self._delete_join_threads.append(Thread(target=task_widget.join, kwargs={'force_join': True}))
        self._delete_join_threads[-1].start()

        self.__created_readuct_widgets_layout.removeWidget(task_widget)
        task_widget.setAttribute(Qt.WA_DeleteOnClose)
        task_widget.close()
        task_widget.setParent(None)  # type: ignore
        self._created_readuct_widgets_holder.updateGeometry()

        return idx

    def _start_stop_all(self, start: bool, force_join: bool = False) -> None:
        if not self.created_readuct_task_widgets:
            write_error_message("You have not created any tasks, yet")
            return
        if self.is_blocked:
            write_error_message("Currently in the process of stopping all tasks, please be patient")
            return
        if start and force_join:
            write_error_message("Internal error, conflicting inputs")
            return
        button = self._play if start else self._stop
        button.setChecked(True)
        if not start:
            for widget in self.created_readuct_task_widgets:
                widget.start_stop(stop_all=True)
            # additional loop for shutdown to collect processes
            self.is_blocked = True
            self._joiner = TaskJoiner(self, force_join)
            self._joiner.start()
        else:
            self._runner = RunAllTasks(self)

            def handle_question(x):
                ans = yes_or_no_question(self, x)
                self._runner.answer_signal.emit(ans)

            self._runner.question_signal.connect(handle_question)  # pylint: disable=no-member
            self._runner.error_signal.connect(write_error_message)  # pylint: disable=no-member
            self._runner.start()
        button.setChecked(False)

    def _load_from_yaml(self, filename: str) -> None:
        systems, names, settings = readuct.load_yaml(filename)
        for name, (method_family, program, calc) in systems.items():
            self._calc_container.add_item(method_family, program, calc.structure, calc.settings.as_dict(), name)

        for (name, inputs, outputs), task_settings in zip(names, settings):
            task_settings['output'] = outputs
            new_widget = TaskWidget(
                parent=self,
                inputs=inputs,
                task_name=CreateReaductTaskWidget.task_mapping[name][0],
                widget_title=name,
                task_settings=task_settings,
                settings_suggestions=None,
            )
            self.add_widget(new_widget)

    def _load_file(self) -> None:
        filename = get_load_file_name(self, "ReaDuctSession", ["yaml", "yml"])
        if filename is None:
            return
        self._load_from_yaml(str(filename))

    def _save_to_yaml(self, filename: str) -> None:
        input_dir = path.join(getcwd(), self.input_file_dir_name)
        aug_systems = self._calc_container.get_augmented_systems()
        if len(aug_systems.keys()) == 0:
            write_error_message("No systems have been created")
            return
        if len(self.created_readuct_task_widgets) == 0:
            write_error_message("No tasks have been created")
            return
        task_names = [task.get_names() for task in self.created_readuct_task_widgets]
        task_settings = [task.get_settings() for task in self.created_readuct_task_widgets]
        if len(task_names) != len(task_settings):
            write_error_message("Could not read all task names and settings")
            return
        # write to calculators to files
        if not path.exists(input_dir):
            mkdir(input_dir)
        for name, (method_family, program, calc) in aug_systems.items():
            su.io.write(str(path.join(input_dir, name)) + ".xyz", calc.structure)
        with open(filename, "w") as f:
            f.write("systems:\n")
            for name, (method_family, program, calc) in aug_systems.items():
                system_file_name = str(path.join(input_dir, name)) + ".xyz"
                f.write(f"  - name: {name}\n")
                f.write(f"    path: {system_file_name}\n")
                f.write(f"    program: {program}\n")
                f.write(f"    method_family: {method_family}\n")
                f.write("    settings:\n")
                for key, value in calc.settings.items():
                    if isinstance(value, str) and not value:
                        f.write(f"      {key}: ''\n")
                    else:
                        f.write(f"      {key}: {value}\n")
                f.write("\n")
            f.write("tasks:\n")
            for (name, inputs, outputs), settings in zip(task_names, task_settings):
                f.write(f"  - type: {name}\n")
                f.write(f"    input: {inputs}\n")
                f.write(f"    output: {outputs}\n")
                if "output" in settings:
                    del settings['output']
                if settings:
                    f.write("    settings:\n")
                    for key, value in settings.items():
                        if isinstance(value, str) and not value:
                            f.write(f"      {key}: ''\n")
                        else:
                            f.write(f"      {key}: {value}\n")

    def _save_file(self) -> None:
        if not self.created_readuct_task_widgets:
            write_error_message("You have not created any tasks, yet")
            return
        filename = get_save_file_name(self, "ReaDuctSession", ["yaml", "yml"])
        if filename is None:
            return
        write_info_message("Stopping all tasks, this may take some time...", timer=1_000_000_000)
        self._start_stop_all(start=False, force_join=True)
        # wait for all tasks to be joined
        while self.is_blocked:
            pass
        clear_status_bar()
        write_info_message("Writing to disk")
        self._save_to_yaml(str(filename))

    def add_system_from_molecular_viewer(self):
        self._calc_container.molecular_viewer_add_button.clicked.emit()  # pylint: disable=no-member


class RunAllTasks(QThread):

    question_signal = Signal(str)
    answer_signal = Signal(bool)
    error_signal = Signal(str)

    def __init__(self, parent: ReaductTab):
        super().__init__(parent)
        self.container = parent
        self._answer: Optional[bool] = None

    def _stop_waiting(self, answer: bool) -> None:
        self._answer = answer
        self.answer_signal.disconnect(self._stop_waiting)

    def _wait_for_answer(self, widget: TaskWidget) -> None:
        self._answer = None
        self.answer_signal.connect(self._stop_waiting)
        self.question_signal.emit(f"The task '{widget.name}' is already finished. "
                                  f"Do you want to re-run it")
        while self._answer is None:
            pass

    def run(self):
        for widget in self.container.created_readuct_task_widgets:
            if not widget.is_working():
                if widget.get_result()[1]:
                    self._wait_for_answer(widget)
                    if not self._answer:
                        # user said no, so we skip this task
                        continue
                messages = []
                widget.start_stop(start_all=True, message_container=messages)
                for message in messages:
                    self.error_signal.emit(message)
            systems, success = widget.join(force_join=False)
            if success:
                print("Updating ", widget.inputs, widget.outputs, systems.keys())
                widget.update_systems.emit(widget.inputs, widget.outputs, systems)
                sleep(0.1)
            else:
                self.error_signal.emit(f"Task '{widget.name}' failed")
            color = widget.green if success else widget.red
            widget.change_color(color)
        self.exit(0)


class TaskJoiner(QThread):

    def __init__(self, parent: ReaductTab, force_join: bool = False):
        super().__init__(parent)
        self.container = parent
        self._force = force_join

    def run(self):
        for widget in self.container.created_readuct_task_widgets:
            widget.join(self._force)
        self.container.is_blocked = False
        self.exit(0)
